"""
Submodule with main function to call to launch batch mode.
"""

import os
import FreeSimpleGUI as sg

from .utils import get_elmt_from_key, create_map, call_auto_map
from .pipeline import batch_pipeline
from .update import update_detection_tab, update_map_tab, update_master_parameters, update_segmentation_tab, update_output_tab
from .input import load, extract_files
from .integrity import sanity_check, check_channel_map_integrity, check_detection_parameters, check_segmentation_parameters, check_output_parameters
from ..gui.layout import _segmentation_layout, _detection_layout, _input_parameters_layout, _ask_channel_map_layout


def batch_promp(
        results_df,
        cell_results_df,
        acquisition_id,
        preset: dict={},
) :

    files_values = [[]]


    #LOAD FILES
    files_table = sg.Table(values=files_values, headings=['Filenames'], col_widths=100, max_col_width= 200, def_col_width=100, num_rows= 10, auto_size_columns=False)

    #DIMENSION SANITY
    sanity_progress = sg.ProgressBar(10, size_px=(500,10), border_width=2)
    sanity_check_button = sg.Button(
        'Check', 
        tooltip= "Will check that all files loaded have the same dimension number and that small fish is able to open them.", 
        pad=(10,0))
    sanity_header = sg.Text("Dimension sanity", font=('bold',15), pad=(0,10))
    dimension_number_text = sg.Text("Dimension number : unknown")
    
#LAYOUT INIT
#########################################
#####   COLUMNS
#########################################

#####   Tabs

    #Input tab
    input_layout = _input_parameters_layout(
        ask_for_segmentation=True,
        is_3D_stack_preset= preset.setdefault("is_3D_stack" ,False),
        time_stack_preset=False,
        multichannel_preset=preset.setdefault("is_multichannel" ,False),
        do_dense_regions_deconvolution_preset=preset.setdefault("do_dense_regions_deconvolution" ,False),
        do_clustering_preset= preset.setdefault("do_cluster_computation", False),
        do_Napari_correction=False,
        do_segmentation_preset= preset.setdefault("segmentation_done", False),
    )
    input_layout += [[sg.Button('Ok')]]
    input_tab = sg.Tab("Input", input_layout)

    napari_correction_elmt = get_elmt_from_key(input_tab, key='show_napari_corrector')

    #Maptab
    map_layout = _ask_channel_map_layout(
        shape=(0,1,2,3,4),
        is_3D_stack=True,
        is_time_stack=True,
        multichannel=True,
    )
    last_shape_read = sg.Text("Last shape read : None")
    auto_map = sg.Button("auto-map", disabled=True, pad=(10,0))
    apply_map_button = sg.Button("apply", disabled=True, pad=(10,0), key='apply-map')
    map_layout += [[last_shape_read]]
    map_layout += [[auto_map, apply_map_button]]
    map_tab = sg.Tab("Map", map_layout)

    #Segmentation tab
    segmentation_layout = _segmentation_layout(
        multichannel=True, 
        cytoplasm_model_preset=preset.setdefault("cyto_model_name",'cyto3'),
        cytoplasm_channel_preset=preset.setdefault("cytoplasm channel",0),
        cyto_diameter_preset=preset.setdefault("cytoplasm diameter",180),
        nucleus_model_preset=preset.setdefault("nucleus_model_name",'nuclei'),
        nucleus_channel_preset=preset.setdefault("nucleus channel",0),
        nucleus_diameter_preset=preset.setdefault("nucleus diameter",150),
        segment_only_nuclei_preset=preset.setdefault("Segment only nuclei",False)
        )
    
    apply_segmentation_button = sg.Button('apply', key='apply-segmentation')
    segmentation_layout += [[apply_segmentation_button]]
    seg_keys_to_hide = ['show segmentation', 'saving path', 'filename', 'other_nucleus_image']
    segmentation_tab = sg.Tab("Segmentation", segmentation_layout, visible=False)

    #Detection tab
    detection_layout = _detection_layout(
        is_3D_stack=True,
        is_multichannel=True,
        do_clustering=True,
        do_dense_region_deconvolution=True,
        do_segmentation=True,
        default_dict= preset,
    )
    apply_detection_button = sg.Button('apply', key='apply-detection')
    detection_layout += [[apply_detection_button]]
    detection_keys_to_hide = ['do_spots_csv', 'do_spots_excel', 'do_spots_feather','spots_filename','spots_extraction_folder']
    detection_tab = sg.Tab("Detection", detection_layout, visible=False)

    #Output tab
    show_batch_folder_text = sg.Text('', key= 'batch_folder_text')
    apply_output_button = sg.Button('apply', key='apply-output')
    save_segmentation_box = sg.Checkbox("save segmentation", disabled=True, key='save segmentation')
    save_detection_box = sg.Checkbox("create spot detection visuals", key= 'save detection', tooltip="Create multichannel tiff with raw spot signal and detected spots.\nWarning if processing a lot of files make sure you have enough free space on your hard drive.")
    extract_spots_box = sg.Checkbox("extract spots", key='extract spots')
    batch_name_input = sg.InputText(size=25, key='batch_name')
    output_layout=[
        [sg.Text("Output folder", font=('bold',15), pad=(0,10))],
        [show_batch_folder_text],
        [sg.Text("Select a folder : "), sg.FolderBrowse(initial_folder=os.getcwd(), key='output_folder', target=(1,-1))],
        [sg.Text("Name for batch : "), batch_name_input],
        [save_segmentation_box],
        [save_detection_box],
        [extract_spots_box],
        [sg.Text("Data extension", font=('bold',15), pad=(0,10))],
        [sg.Checkbox(".csv", key='csv'),sg.Checkbox(".xlsx", key='xlsx'),sg.Checkbox(".feather", key='feather'),],
        [apply_output_button],
    ]
    output_tab = sg.Tab("Output", output_layout, visible=True)

    ##TAB GROUP
    _tab_group = sg.TabGroup([[input_tab, map_tab, segmentation_tab, detection_tab, output_tab]], enable_events=True)
    tab_col = sg.Column( #Allow the tab to be scrollable
        [[_tab_group]],
        scrollable=True,
        vertical_scroll_only=True,
        s= (390,390),
        pad=((0,0),(5,5))
        )
    
    tab_dict= {
        "Input" : input_tab,
        "Segmentation" : segmentation_tab,
        "Detection" : detection_tab,
        "Map" : map_tab,
        "Output" : output_tab
    }
    
#####   Launcher

    start_button =sg.Button('Start', button_color= 'green', disabled= True, pad= ((115,5),(0,10)))
    stop_button = sg.Button('Cancel', button_color= 'red', pad= ((5,115),(0,10)))
    batch_progression_bar = sg.ProgressBar(max_value=0, size_px=(340,20), bar_color= ('blue','black'), border_width=2, pad=((15,0),(30,0)))
    mapping_ok_text = sg.Text('Uncorrect mapping', text_color='gray', font='roman 14 bold', pad=((0,0),(100,5)))
    segmentation_ok_text = sg.Text('Uncorrect segmentation settings', font='roman 14 bold', pad=(0,5), text_color='gray', visible=False)
    detection_ok_text = sg.Text('Uncorrect detection settings', text_color='gray', font='roman 14 bold', pad=(0,5))
    output_ok_text = sg.Text('Uncorrect output parameters', text_color='gray', font='roman 14 bold', pad=(0,5))
    current_acquisition_text = sg.Text('0', text_color='gray', font='roman 15 bold', pad= ((150,5),(10,10)))
    total_acquisition_text = sg.Text('/ 0', text_color='gray', font='roman 15 bold', pad= ((5,150),(10,10)))

    launcher_layout = [
        [mapping_ok_text],
        [segmentation_ok_text],
        [detection_ok_text],
        [output_ok_text],
        [batch_progression_bar],
        [current_acquisition_text,total_acquisition_text],
        [start_button, stop_button],
    ]
    launch_col = sg.Column( #Allow the tab to be scrollable
        launcher_layout,
        s= (390,390),
        pad=((3,5),(5,5))
        )

#########################################
#####   Window Creation
#########################################
    stream_output = sg.Output(size=(100,10), pad=(30,10), visible=True)
    layout = [
        [sg.Text("Batch Processing", font=('bold',20), pad=((300,0),(0,2)))],
        [sg.Text("Select a folder : "), sg.FolderBrowse(initial_folder=os.getcwd(), key='Batch_folder'), sg.Button('Load')],
        [files_table],
        [sanity_header, sanity_check_button, sanity_progress],
        [dimension_number_text],
        [tab_col, launch_col],
        [stream_output],
    ]

    window = sg.Window("small fish", layout=layout, size= (800,800), auto_size_buttons=True, auto_size_text=True, resizable=True)
    
    #MASTER PARAMETERS
    Master_parameters_dict ={
        '_map' : {},
        '_is_mapping_correct' : False,
        '_is_segmentation_correct' : None, # None : segmentation disabled; Then true/false if enabled
        '_is_detection_correct' : False,
        '_is_output_correct' : False,
    }
    Master_parameters_update_dict = {
        '_is_mapping_correct' : mapping_ok_text,
        '_is_segmentation_correct' : segmentation_ok_text, # None : segmentation disabled; Then true/false if enabled
        '_is_detection_correct' : detection_ok_text,
        '_is_output_correct' : output_ok_text,

    }
    loop = 0
    timeout = 1
    last_shape = None
    talk=True
    
#########################################
#####   Event Loop : break to close window
#########################################
    
    #Hiding options for non batch mode
    window= window.finalize()
    napari_correction_elmt.update(disabled=True)
    get_elmt_from_key(tab_dict['Input'], key= 'image_path').update(disabled=True)
    for key in seg_keys_to_hide : get_elmt_from_key(tab_dict['Segmentation'], key=key).update(disabled=True)
    for key in detection_keys_to_hide : get_elmt_from_key(tab_dict['Detection'], key=key).update(disabled=True)
    
    while True : 
        try :
            loop +=1
            window = window.refresh()
            event, values = window.read(timeout=timeout)

            #Welcome message
            if loop == 1 : 
                timeout = 500
                print("Welcome to small fish batch analysis. Please start by loading some files and setting parameters.")

            batch_folder = values.get('Batch_folder')
            is_multichanel = values.get('is_multichannel')
            is_3D = values.get('is_3D_stack')
            do_segmentation = values.get('do_segmentation')
            do_dense_regions_deconvolution = values.get('do_dense_regions_deconvolution')
            do_clustering = values.get('do_cluster_computation')

            if type(batch_folder) != type(None)  and event == 'Load':

                files_values, last_shape, dim_number = load(batch_folder)
                files_table.update(values=files_values)
                last_shape_read.update("Last shape read : {0}".format(last_shape))
                dimension_number_text.update("Dimension number : {0}".format(dim_number))
                update_map_tab(
                    tab_elmt=tab_dict.get("Map"),
                    is_3D=is_3D,
                    is_multichannel=is_multichanel,
                    last_shape=last_shape
                )
                filename_list = extract_files(files_values)

            elif event == 'Check' :
                last_shape = sanity_check(
                    filename_list=filename_list,
                    batch_folder=batch_folder,
                    window=window,
                    progress_bar=sanity_progress
                )
                if isinstance(last_shape,(tuple,list)) :
                    dim_number = len(last_shape)
                    dimension_number_text.update("Dimension number : {0}".format(dim_number))
                    auto_map.update(disabled=False)
                else :
                    dim_number = None
                    dimension_number_text.update("Dimension number : unknown")
                    auto_map.update(disabled=True)

                last_shape_read.update("Last shape read : {0}".format(last_shape))

            elif event == _tab_group.key or event == 'Ok': #Tab switch in parameters

                update_map_tab(
                    tab_elmt=tab_dict.get("Map"),
                    is_3D=is_3D,
                    is_multichannel=is_multichanel,
                    last_shape=last_shape
                )

            elif event == 'auto-map' :
                Master_parameters_dict['_map'] = call_auto_map(
                    tab_elmt=tab_dict.get("Map"),
                    shape=last_shape,
                    is_3D=is_3D,
                    is_multichannel=is_multichanel
                )

                Master_parameters_dict['_is_mapping_correct'] = check_channel_map_integrity(
                    maping=Master_parameters_dict['_map'],
                    shape=last_shape,
                    expected_dim=dim_number
                    )

                if not Master_parameters_dict['_is_mapping_correct'] : Master_parameters_dict['_map'] = {}

            elif event == 'apply-map' :

                Master_parameters_dict['_map'] = create_map(
                    values=values,
                    is_3D=is_3D,
                    is_multichannel=is_multichanel,
                )

                Master_parameters_dict['_is_mapping_correct'] = check_channel_map_integrity(
                    maping=Master_parameters_dict['_map'],
                    shape=last_shape,
                    expected_dim=dim_number
                    )

                if not Master_parameters_dict['_is_mapping_correct'] : Master_parameters_dict['_map'] = {}

            elif event == 'apply-segmentation' :
                Master_parameters_dict['_is_segmentation_correct'], values = check_segmentation_parameters(
                    values=values,
                    shape=last_shape,
                    is_multichannel=is_multichanel
                )

            elif event == 'apply-detection' :
                Master_parameters_dict['_is_detection_correct'], values = check_detection_parameters(
                    values=values,
                    do_dense_region_deconvolution=do_dense_regions_deconvolution,
                    do_clustering=do_clustering,
                    is_multichannel=is_multichanel,
                    is_3D=is_3D,
                    map_= Master_parameters_dict.get('_map'),
                    shape=last_shape
                )

            elif event == 'apply-output' :
                Master_parameters_dict['_is_output_correct'], values = check_output_parameters(values)
                batch_name_input.update(value=values.get('batch_name'))

            elif event == "Cancel" :
                stream_output.restore_stderr()
                stream_output.restore_stdout()
                window.close()
                return results_df, cell_results_df, acquisition_id, preset, False, None,None #Segmentation done : False, cell label : None, Nucleus label : None


            elif event == None :
                raise InterruptedError("closed")
            elif event == sg.WINDOW_CLOSED : 
                raise InterruptedError("closed")

            elif event == 'Start' :
                start_button.update(disabled=True)
                results_df, cell_results_df, acquisition_id = batch_pipeline(
                    batch_window= window,
                    batch_progress_bar= batch_progression_bar,
                    progress_count=current_acquisition_text,
                    parameters=values,
                    filenames_list=filename_list,
                    do_segmentation=do_segmentation,
                    map_= Master_parameters_dict['_map'],
                    results_df=results_df,
                    cell_results_df=cell_results_df,
                    is_3D=is_3D,
                    last_acquisition_id=acquisition_id+1,
                )
                stream_output.restore_stderr()
                stream_output.restore_stdout()

                window.close()
                return results_df, cell_results_df, acquisition_id, values, False, None,None #Segmentation done : False, cell label : None, Nucleus label : None


            #End of loop

            #If user loads other files
            if type(last_shape) != type(None) :
                if len(last_shape) != 2 + is_3D + is_multichanel :
                    Master_parameters_dict['_is_mapping_correct'] = False
                    Master_parameters_dict['_is_detection_correct'] = False
                    Master_parameters_dict['_is_segmentation_correct'] = False
            else : 
                Master_parameters_dict['_is_mapping_correct'] = False
                Master_parameters_dict['_is_detection_correct'] = False
                Master_parameters_dict['_is_segmentation_correct'] = False

            update_master_parameters(
                Master_parameter_dict=Master_parameters_dict,
                update_dict=Master_parameters_update_dict
            )

            update_segmentation_tab(
                tab_elmt=tab_dict.get("Segmentation"),
                segmentation_correct_text= segmentation_ok_text,
                do_segmentation=do_segmentation,
                is_multichannel=is_multichanel,
                is_mapping_ok=Master_parameters_dict['_is_mapping_correct']
            )

            update_detection_tab(
                tab_elmt=tab_dict.get("Detection"),
                is_multichannel=is_multichanel,
                is_3D=is_3D,
                do_dense_region_deconvolution=do_dense_regions_deconvolution,
                do_clustering=do_clustering,
                is_mapping_ok=Master_parameters_dict['_is_mapping_correct'],
                do_segmentation=do_segmentation,
            )

            update_output_tab(
                tab_elmt=tab_dict['Output'],
                do_segmentation=do_segmentation,
                output_folder=values['output_folder'],
            )

            _segmentation_ready = (Master_parameters_dict['_is_segmentation_correct'] and do_segmentation) or not do_segmentation
            if Master_parameters_dict['_is_detection_correct'] and Master_parameters_dict['_is_output_correct'] and Master_parameters_dict['_is_mapping_correct'] and _segmentation_ready :
                start_button.update(button_color='blue', disabled=False)
                total_acquisition_text.update("/ {0}".format(len(filename_list)))
                batch_progression_bar.update(current_count=0, bar_color=('green','blue'))
                if talk : 
                    print("Small fish is ready, launch with the start button.")
                    talk= False
            else :
                talk = True
                start_button.update(button_color='green', disabled=True)
                total_acquisition_text.update("/ 0")
                batch_progression_bar.update(current_count=0, bar_color=('blue','black'))

        except Exception as e :
            stream_output.restore_stderr()
            stream_output.restore_stdout()
            raise e


    window.close()


from bigfish.stack import read_image
from czifile import imread
from ..utils import check_parameter
import re
from aicsimageio import AICSImage
from typing import Optional, Tuple

class FormatError(Exception):
    pass


def open_image(full_path:str) :
    if full_path.endswith('.czi') : im = imread(full_path)
    else : im = read_image(full_path)

    reshape = []
    for axis in im.shape :
        if axis != 1 : reshape.append(axis)
    im = im.reshape(reshape)

    return im


def check_format(image, is_3D, is_multichannel) :
    shape = list(image.shape)
    dim = image.ndim - (shape[image.ndim - 1] == 1)
    if not dim == (2 + is_3D  + is_multichannel) :
        raise FormatError("Inconsistency in image format and parameters.")



def get_filename(full_path: str) :
    check_parameter(full_path=str)

    pattern = r'.*\/(.+)\..*$'
    if not full_path.startswith('/') : full_path = '/' + full_path
    re_match = re.findall(pattern, full_path)
    if len(re_match) == 0 : raise ValueError("Could not read filename from image full path.")
    if len(re_match) == 1 : return re_match[0]
    else : raise AssertionError("Several filenames read from path")

def get_voxel_size(filepath: str) -> Optional[Tuple[Optional[float], Optional[float], Optional[float]]]:
    """
    Returns voxel size in nanometers (nm) as a tuple (X, Y, Z).
    Any of the dimensions may be None if not available.
    /WARINING\ : the unit might not be nm
    """
    try:
        img = AICSImage(filepath)
        voxel_sizes = img.physical_pixel_sizes  # values in meters
        if voxel_sizes is None:
            return None
        x = voxel_sizes.X * 1e3 if voxel_sizes.X else None
        y = voxel_sizes.Y * 1e3 if voxel_sizes.Y else None
        z = voxel_sizes.Z * 1e3 if voxel_sizes.Z else None
        return (z, y, x)
    except Exception as e:
        print(f"Failed to read voxel size from {filepath}: {e}")
        return None


import FreeSimpleGUI as sg
import small_fish_gui.batch as batch
import pandas as pd

try :
    batch.batch_promp(pd.DataFrame(), pd.DataFrame(), acquisition_id=0,)
except Exception as e :
    raise e

print('end')


import os
from typing import Optional, Tuple

IMAGE_PATH = "/home/floric/Documents/python/small_fish/images_test/Cropped_image_dapi_0.tif"

def get_voxel_size_0(filepath: str) -> Optional[Tuple[float, float, float]]:
    ext = os.path.splitext(filepath)[1].lower()
    
    if ext in ['.tif', '.tiff']:
        import tifffile
        with tifffile.TiffFile(filepath) as tif:
            tags = tif.pages[0].tags
            print(tags)
            x_res = tags.get('XResolution')
            y_res = tags.get('YResolution')
            if x_res and y_res:
                # Resolution is typically in (numerator, denominator)
                x = x_res.value[1] / x_res.value[0]
                y = y_res.value[1] / y_res.value[0]
                # Microscopy images may store pixel size in microns per pixel (invert if necessary)
                return (1.0, y, x)  # Z usually not in TIFF metadata, so set as 1.0 or None
            return None
    
    elif ext == '.czi':
        import czifile
        with czifile.CziFile(filepath) as czi:
            metadata = czi.metadata(raw=False)
            scaling = metadata.get('ImageDocument', {}).get('Metadata', {}).get('Scaling', {})
            try:
                x = float(scaling['Items']['Distance'][0]['Value']) * 1e6  # in microns
                y = float(scaling['Items']['Distance'][1]['Value']) * 1e6
                z = float(scaling['Items']['Distance'][2]['Value']) * 1e6 if len(scaling['Items']['Distance']) > 2 else 1.0
                return (x, y, z)
            except Exception:
                return None
    
    return None

from aicsimageio import AICSImage
from typing import Optional, Tuple

def get_voxel_size(filepath: str) -> Optional[Tuple[Optional[float], Optional[float], Optional[float]]]:
    """
    Returns voxel size in nanometers (nm) as a tuple (X, Y, Z).
    Any of the dimensions may be None if not available.
    /WARINING\ : the unit might not be nm
    """
    try:
        img = AICSImage(filepath)
        voxel_sizes = img.physical_pixel_sizes  # values in meters
        if voxel_sizes is None:
            return None
        x = voxel_sizes.X * 1e3 if voxel_sizes.X else None
        y = voxel_sizes.Y * 1e3 if voxel_sizes.Y else None
        z = voxel_sizes.Z * 1e3 if voxel_sizes.Z else None
        return (x, y, z)
    except Exception as e:
        print(f"Failed to read voxel size from {filepath}: {e}")
        return None


res = get_voxel_size("/home/floric/Documents/python/small_fish/images_test/Cropped_image_dapi_0.tif")
print("RES : ",res)



"""
This submodule groups all the possible actions of the user in the main windows. It is the start of each action the user can do.
"""

from ..gui.prompts import output_image_prompt, prompt_save_segmentation, prompt_load_segmentation
from ..gui.prompts import ask_detection_confirmation, ask_cancel_detection, ask_confirmation
from ..gui.prompts import rename_prompt

from ..interface.inoutput import write_results, write_list_of_results
from ..interface.inoutput import input_segmentation, output_segmentation

from ._preprocess import map_channels
from ._preprocess import prepare_image_detection
from ._preprocess import reorder_shape, reorder_image_stack
from ._preprocess import ask_input_parameters

from .detection import initiate_detection, launch_detection, launch_features_computation
from .detection import get_nucleus_signal
from .spots import launch_spots_extraction

from .segmentation import launch_segmentation
from ._colocalisation import initiate_colocalisation, launch_colocalisation

from ..hints import pipeline_parameters
from ..__init__ import __wiki__

import os
import pandas as pd
import FreeSimpleGUI as sg
import numpy as np
import webbrowser

def open_wiki() :
    webbrowser.open_new_tab(__wiki__)

def segment_cells(user_parameters : pipeline_parameters, nucleus_label, cytoplasm_label) :
    if user_parameters['segmentation_done'] :
        if ask_confirmation("A segmentation is in small fish memory, do you want to erase it ? (you will not be able to undo)") :
            user_parameters['segmentation_done'] = False
            nucleus_label = None
            cytoplasm_label = None
        else : 
            return nucleus_label, cytoplasm_label, user_parameters

    nucleus_label, cytoplasm_label, user_parameters = launch_segmentation(
        user_parameters,
        nucleus_label=nucleus_label,
        cytoplasm_label=cytoplasm_label,
    )

    if type(cytoplasm_label) != type(None) and type(nucleus_label) != type(None) :
        user_parameters['segmentation_done'] = True

    return nucleus_label, cytoplasm_label, user_parameters

def add_detection(user_parameters : pipeline_parameters, acquisition_id, cytoplasm_label, nucleus_label) :

    new_results_df = pd.DataFrame()
    new_cell_results_df = pd.DataFrame()

    #Ask for image parameters
    new_parameters = ask_input_parameters(ask_for_segmentation= False) #The image is open and stored inside user_parameters
    if type(new_parameters) == type(None) : #if user clicks 'Cancel'
        return new_results_df, new_cell_results_df, acquisition_id, user_parameters
    else :
        user_parameters.update(new_parameters)

    map_ = map_channels(user_parameters)
    if type(map_) == type(None) : #User clicks Cancel 
        return new_results_df, new_cell_results_df, acquisition_id, user_parameters
    user_parameters['reordered_shape'] = reorder_shape(user_parameters['shape'], map_)

    #Detection
    while True : # This loop allow user to try detection with different thresholds or parameters before launching features computation
        detection_parameters = initiate_detection(
            user_parameters,
            map_= map_,
            shape = user_parameters['image'].shape
            )

        if type(detection_parameters) != type(None) :
            user_parameters.update(detection_parameters) 
        else : #If user clicks cancel
            
            cancel = ask_cancel_detection()
            if cancel : 
                return new_results_df, new_cell_results_df, acquisition_id, user_parameters
            else : continue

        acquisition_id += 1
        image, other_image = prepare_image_detection(map_, user_parameters) 
        nucleus_signal = get_nucleus_signal(image, other_image, user_parameters)
        
        try : # Catch error raised if user enter a spot size too small compare to voxel size
            user_parameters, frame_result, spots, clusters, spots_cluster_id = launch_detection(
                image,
                other_image,
                user_parameters,
                cell_label=cytoplasm_label,
                nucleus_label=nucleus_label
            )

        except ValueError as error :
            if "The array should have an upper bound of 1" in str(error) :
                sg.popup("Spot size too small for current voxel size.")
                continue
            else :
                raise(error)


        if user_parameters['show_napari_corrector'] or user_parameters['show_interactive_threshold_selector']:
            if ask_detection_confirmation(user_parameters.get('threshold')) : break
        else :
            break

    if user_parameters['spots_extraction_folder'] != '' and type(user_parameters['spots_extraction_folder']) != type(None) :
        if user_parameters['spots_filename'] != '' and type(user_parameters['spots_filename']) != type(None) :
            if any((user_parameters['do_spots_excel'], user_parameters['do_spots_csv'], user_parameters['do_spots_feather'])) :
                launch_spots_extraction(
                    acquisition_id=acquisition_id,
                    user_parameters=user_parameters,
                    image=image,
                    spots=spots,
                    cluster_id= spots_cluster_id,
                    nucleus_label= nucleus_label,
                    cell_label= cytoplasm_label,
                )

    #Features computation
    new_results_df, new_cell_results_df = launch_features_computation(
    acquisition_id=acquisition_id,
    image=image,
    nucleus_signal = nucleus_signal,
    spots=spots,
    clusters=clusters,
    spots_cluster_id = spots_cluster_id,
    nucleus_label = nucleus_label,
    cell_label= cytoplasm_label,
    user_parameters=user_parameters,
    frame_results=frame_result,
    )
    return new_results_df, new_cell_results_df, acquisition_id, user_parameters

def save_segmentation(nucleus_label : np.ndarray, cytoplasm_label: np.ndarray) :
    if type(nucleus_label) == type(None) or type(cytoplasm_label) == type(None) :
        sg.popup("No segmentation to save.")
    
    else :

        while True :
            answer = prompt_save_segmentation()
            if type(answer) == type(None) : 
                return False #User clicks cancel

            path = answer['folder'] + '/' + answer['filename']
            is_npy, is_npz, is_npz_compressed = answer['ext'], answer['ext0'], answer['ext1']

            if is_npy + is_npz + is_npz_compressed == 1 : 
                if is_npy : 
                    extension = 'npy'
                    if os.path.isfile(path + '_nucleus_segmentation.npy') or os.path.isfile(path + '_cytoplasm_segmentation.npy') : 
                        if ask_confirmation("File exists. Replace ?") :
                            break
                        else :
                            pass
                    else :
                        break

                elif is_npz :
                    extension = 'npz_uncompressed'
                    if os.path.isfile(path + '_nucleus_segmentation.npz') or os.path.isfile(path + '_cytoplasm_segmentation.npz') : 
                        if ask_confirmation("File exists. Replace ?") :
                            break
                        else :
                            pass
                    else :
                        break

                elif is_npz_compressed : 
                    extension = 'npz_compressed'
                    if os.path.isfile(path + '_nucleus_segmentation.npz') or os.path.isfile(path + '_cytoplasm_segmentation.npz') : 
                        if ask_confirmation("File exists. Replace ?") :
                            break
                        else :
                            pass
                    else :
                        break

            else :
                sg.popup("Please select an extension.")

        saved = output_segmentation(
            path,
            extension,
            nucleus_label,
            cytoplasm_label,
            )
        if saved : sg.popup("Segmentation was saved at {0}.".format(path))
        else : sg.popup("No segmentation was saved..")
        return True
    
def load_segmentation(nucleus_label, cytoplasm_label, segmentation_done) :

    if segmentation_done :
        if ask_confirmation("Segmentation already in memory. Replace ?") :
            pass
        else :
            return nucleus_label, cytoplasm_label, segmentation_done

    while True :
        answer = prompt_load_segmentation()

        if type(answer) == type(None) : #user clicks cancel
            return nucleus_label, cytoplasm_label, segmentation_done

        try :
            nucleus_label, cytoplasm_label = input_segmentation(
                answer['nucleus'],
                answer['cytoplasm'],
            )
        
        except ValueError as e :
            sg.popup(str(e))
        else :
            break

    if type(nucleus_label) != type(None) and type(nucleus_label) != np.ndarray :
        nucleus_label = nucleus_label['arr_0']

    if type(cytoplasm_label) != type(None) and type(cytoplasm_label) != np.ndarray :
        cytoplasm_label = cytoplasm_label['arr_0']

    segmentation_done = (type(nucleus_label) != type(None) and type(cytoplasm_label) != type(None))

    if segmentation_done : assert type(nucleus_label) == np.ndarray and type(cytoplasm_label) == np.ndarray

    return nucleus_label, cytoplasm_label, segmentation_done

def save_results(
        result_df : pd.DataFrame, 
        cell_result_df : pd.DataFrame, 
        global_coloc_df : pd.DataFrame, 
        cell_coloc_df : dict, #TODO : Rename to cell_coloc_dict
        ) :
    if len(result_df) != 0 :
        dic = output_image_prompt(filename=result_df.iloc[0].at['filename'])

        if isinstance(dic, dict) :
            path = dic['folder']
            filename = dic['filename']
            do_excel = dic['Excel']
            do_feather = dic['Feather']
            do_csv = dic['csv']

            if 'rna_coords' in cell_result_df.columns : cell_result_df = cell_result_df.drop(columns='rna_coords')

            sucess1 = write_results(result_df, path= path, filename=filename, do_excel= do_excel, do_feather= do_feather, do_csv=do_csv)
            sucess2 = write_results(cell_result_df, path= path, filename=filename + '_cell_result', do_excel= do_excel, do_feather= do_feather, do_csv=do_csv)
            sucess3 = write_results(global_coloc_df, path= path, filename=filename + 'global_coloc_result', do_excel= do_excel, do_feather= do_feather, do_csv=do_csv)
            sucess4 = write_list_of_results(cell_coloc_df.values(), path= path, filename=filename + 'cell2cell_coloc_result', do_excel= do_excel, do_feather= do_feather, do_csv=do_csv)
            if all([sucess1,sucess2, sucess3, sucess4,]) : sg.popup("Sucessfully saved at {0}.".format(path))

    else :
        dic = None
        sg.popup('No results to save.') 

def compute_colocalisation(result_tables, result_dataframe, cell_result_dataframe, global_coloc_df, cell_coloc_df) :
    colocalisation_distance = initiate_colocalisation(result_tables)

    if colocalisation_distance == False :
        pass
    else :
        global_coloc_df, cell_coloc_df = launch_colocalisation(
            result_tables, 
            result_dataframe=result_dataframe, 
            cell_result_dataframe=cell_result_dataframe,
            colocalisation_distance=colocalisation_distance,
            global_coloc_df=global_coloc_df,
            cell_coloc_df=cell_coloc_df,
            )

    return global_coloc_df, cell_coloc_df

def delete_acquisitions(selected_acquisitions : pd.DataFrame, 
                        result_df : pd.DataFrame, 
                        cell_result_df : pd.DataFrame, 
                        global_coloc_df : pd.DataFrame,
                        cell_coloc_df : dict,
                        ) :
    
    if len(result_df) == 0 :
        sg.popup("No acquisition to delete.")
        return result_df, cell_result_df, global_coloc_df

    if len(selected_acquisitions) == 0 :
        sg.popup("Please select the acquisitions you would like to delete.")
    else :
        acquisition_ids = list(result_df.iloc[list(selected_acquisitions)]['acquisition_id'])
        result_drop_idx = result_df[result_df['acquisition_id'].isin(acquisition_ids)].index
        print("{0} acquisitions deleted.".format(len(result_drop_idx)))
        
        if len(cell_result_df) > 0 :
            cell_result_df_drop_idx = cell_result_df[cell_result_df['acquisition_id'].isin(acquisition_ids)].index
            print("{0} cells deleted.".format(len(cell_result_df_drop_idx)))
            cell_result_df = cell_result_df.drop(cell_result_df_drop_idx, axis=0)
        
        if len(global_coloc_df) > 0 :
            coloc_df_drop_idx = global_coloc_df[(global_coloc_df["acquisition_id_1"].isin(acquisition_ids)) | (global_coloc_df['acquisition_id_2'].isin(acquisition_ids))].index
            print("{0} coloc measurement deleted.".format(len(coloc_df_drop_idx)))
            global_coloc_df = global_coloc_df.drop(coloc_df_drop_idx, axis=0)
        
        if len(cell_coloc_df) > 0 :
            keys_to_delete = []
            for acquisition_id in acquisition_ids :
                for coloc_key in cell_coloc_df.keys() :
                    if acquisition_id in coloc_key :
                        keys_to_delete.append(coloc_key) 

            for key in keys_to_delete : 
                if key in cell_coloc_df.keys() : cell_coloc_df.pop(key)

        result_df = result_df.drop(result_drop_idx, axis=0)

    return result_df, cell_result_df, global_coloc_df, cell_coloc_df

def rename_acquisitions(
        selected_acquisitions : pd.DataFrame, 
        result_df : pd.DataFrame, 
        cell_result_df : pd.DataFrame, 
        global_coloc_df : pd.DataFrame,
        cell_coloc_df : dict,
        ) :
    
    if len(result_df) == 0 :
        sg.popup("No acquisition to rename.")
        return result_df, cell_result_df, global_coloc_df, cell_coloc_df
    
    if len(selected_acquisitions) == 0 :
        sg.popup("Please select the acquisitions you would like to rename.")

    else :
        name = rename_prompt()
        if not name : return result_df, cell_result_df, global_coloc_df, cell_coloc_df #User didn't put a name or canceled
        name : str = name.replace(' ','_')
        acquisition_ids = list(result_df.iloc[list(selected_acquisitions)]['acquisition_id'])
        old_names = list(result_df.loc[result_df['acquisition_id'].isin(acquisition_ids)]['name'])
        old_names.sort(key=len) #We order this list by elmt length
        old_names.reverse() #From longer to smaller

        result_df.loc[result_df['acquisition_id'].isin(acquisition_ids),['name']] = name
        if len(cell_result_df) > 0 : cell_result_df.loc[cell_result_df['acquisition_id'].isin(acquisition_ids),['name']] = name
        if len(global_coloc_df) > 0 : 
            global_coloc_df.loc[global_coloc_df['acquisition_id_1'].isin(acquisition_ids), ['name1']] = name
            global_coloc_df.loc[global_coloc_df['acquisition_id_2'].isin(acquisition_ids), ['name2']] = name
        if len(cell_coloc_df) > 0 :
            for key in cell_coloc_df.keys() :
                df = cell_coloc_df[key]
                target_columns = df.columns.get_level_values(1)
                for old_name in old_names : #Note list was ordered by elmt len (decs) to avoid conflict when one name is contained by another one. if the shorter is processed first then the longer will not be able to be properly renamed.
                    target_columns = target_columns.str.replace(old_name, name)

                new_columns = zip(
                    df.columns.get_level_values(0),
                    target_columns,
                    df.columns.get_level_values(2),
                )

                df.columns = pd.MultiIndex.from_tuples(new_columns)
                cell_coloc_df[key] = df

    return result_df, cell_result_df, global_coloc_df, cell_coloc_df


from ._custom_errors import MissMatchError
from ..gui import coloc_prompt, add_default_loading

import numpy as np
import pandas as pd
import FreeSimpleGUI as sg
from scipy.ndimage import distance_transform_edt
from scipy.signal import fftconvolve

def reconstruct_boolean_signal(image_shape, spot_list: list):
    signal = np.zeros(image_shape, dtype= bool)
    if len(spot_list) == 0 : return signal
    dim = len(spot_list[0])

    if dim == 3 :
        Z, Y, X = list(zip(*spot_list))
        signal[Z,Y,X] = True

    else : 
        Y, X = list(zip(*spot_list))
        signal[Y,X] = True
        

    return signal

def nanometer_to_pixel(value, scale) :
    if isinstance(scale, (float,int)) : scale = [scale]
    if isinstance(value, (float,int)) : value = [value]*len(scale)
    if len(value) != len(scale) : raise ValueError("value and scale must have the same dimensionality")

    return list(np.array(value) / np.array(scale))

def _create_counting_kernel(radius_nm, voxel_size) :

    max_pixel_distance = int(max(nanometer_to_pixel(radius_nm, voxel_size)))
    kernel = np.ones(shape=[2*max_pixel_distance+1 for i in range(len(voxel_size))]) #always odd number so middle is always at [pixel_radius-1, pixel_radius-1, pixel_radius-1]
    if len(voxel_size) == 3 :
        kernel[max_pixel_distance, max_pixel_distance, max_pixel_distance] = 0
    else :
        kernel[max_pixel_distance, max_pixel_distance] = 0

    kernel = distance_transform_edt(kernel, sampling= voxel_size) <= radius_nm
    
    return kernel.astype(int)

def _spot_count_map(spots_array, radius_px, voxel_size) :
    """
    Create a map where each pixel value correspond to the number of spots closer than radius to the position.
    """

    kernel = _create_counting_kernel(radius_px, voxel_size)
    map = fftconvolve(spots_array, kernel, mode= 'same')

    return np.round(map).astype(int)

def _reconstruct_spot_signal(image_shape, spot_list: list, dim=3):
    """
    Create a map where each pixel value correspond to the number of spots located in this position.
    """
    signal = np.zeros(image_shape, dtype= int)
    unique_list, counts = np.unique(spot_list, return_counts= True, axis=0)
    if dim == 3 :
        Z, Y, X = list(zip(*unique_list))
        signal[Z,Y,X] = counts
    elif dim == 2 :
        Y, X = list(zip(*unique_list))
        signal[Y,X] = counts
    else : 
        raise ValueError("Wrong dim passed should be 2 or 3, it is {0}".format(dim))

    return signal

def spots_multicolocalisation(spots_list, anchor_list, radius_nm, image_shape, voxel_size) :

    """
    Compute the number of spots from spots_list closer than radius to a spot from anchor_list. Each spots_list spots will be counted as many times as there are anchors close enough.
    Note that the radius in nm is converted to pixel using voxel size, and rounded to the closest int value.
    
    Example in 2D
    --------

    >>> Anchors         spots           Radius (2px)    Count
    >>> 0 0 0 0 0 0     0 X 0 0 X 0       1             0 1 0 0 0 0
    >>> 0 X 0 0 0 0     X 0 0 X 0 0     1 1 1           1 0 0 0 0 0
    >>> 0 X 0 0 0 0     X X 0 0 0 0       1             1 2 0 0 0 0     --> 5
    >>> 0 0 0 0 X 0     0 0 X 0 0 0                     0 0 0 0 0 0
    >>> 0 0 0 0 0 0     0 0 0 X 0 0                     0 0 0 0 0 0

    Parameters
    ----------
    spots_list : list
    anchor_list : list
    radius_nm : int, float
    image_shape : tuple (Z, Y, X)
    voxel_size : tuple (Z, Y, X)
    
    Returns
    -------
    Returns the list of neighbouring spot number to 'spots_list'.
    """
    if len(spots_list) == 0 or len(anchor_list) == 0 : return 0
    if len(voxel_size) != len(spots_list[0]) : raise ValueError("Dimensions missmatched; voxel_size : {0} spots : {1}".format(len(voxel_size), len(spots_list[0])))

    dim = len(voxel_size)

    anchor_array = _reconstruct_spot_signal(image_shape=image_shape, spot_list=anchor_list, dim=dim)
    count_map = _spot_count_map(anchor_array, radius_px=radius_nm, voxel_size=voxel_size)

    if dim == 3 :
        Z,Y,X = list(zip(*spots_list))
        res = list(count_map[Z,Y,X])

    if dim == 2 :
        Y,X = list(zip(*spots_list))
        res = list(count_map[Y,X])

    return res

def spots_colocalisation(spot_list1:list, spot_list2:list, distance: float, voxel_size)-> int :
    """
    Return number of spots from spot_list1 located closer(large) than distance to at least one spot of spot_list2.

    Parameters
    ----------
        image_shape : tuple
        spot_list1 : list
        spot_list2 : list
        distance : nanometer
            distance in nanometer.
        voxel_size : (z,y,x) tuple
    """

    if len(spot_list1) == 0 or len(spot_list2) == 0 : return np.NaN
    if len(spot_list1[0]) != len(spot_list2[0]) : 
        raise MissMatchError("dimensionalities of spots 1 and spots 2 don't match.")
    
    shape1 = np.max(spot_list1,axis=0)
    shape2 = np.max(spot_list2,axis=0)
    image_shape = np.max([shape1, shape2],axis=0) + 1

    signal2 = reconstruct_boolean_signal(image_shape, spot_list2)
    mask = np.logical_not(signal2)
    distance_map = distance_transform_edt(mask, sampling= voxel_size)

    if len(voxel_size) == 3 :
        Z,Y,X = zip(*spot_list1)
        count = (distance_map[Z,Y,X] <= distance).sum()
    else :
        Y,X = zip(*spot_list1)
        count = (distance_map[Y,X] <= distance).sum()

    return count


   



def initiate_colocalisation(result_tables) :
    if len(result_tables) != 2 : 
        sg.popup("Please select 2 acquisitions for colocalisation (Ctrl + click in the table)")
        return False
    else : 
        while True :
            colocalisation_distance = coloc_prompt()
            if colocalisation_distance == False : return False
            try : 
                colocalisation_distance = int(colocalisation_distance)
            except Exception :
                sg.popup("Incorrect colocalisation distance")
            else :
                break
        return colocalisation_distance

def _global_coloc(acquisition_id1,acquisition_id2, result_dataframe, colocalisation_distance) :
    """

    Target :

    - acquisition_couple
    - colocalisation_distance
    - spot1_total
    - spot2_total
    - fraction_spot1_coloc_spots
    - fraction_spot2_coloc_spots
    - fraction_spot1_coloc_clusters
    - fraction_spot2_coloc_spots

    """

    acquisition1 = result_dataframe.loc[result_dataframe['acquisition_id'] == acquisition_id1]
    acquisition2 = result_dataframe.loc[result_dataframe['acquisition_id'] == acquisition_id2]

    assert len(acquisition1) == 1
    assert len(acquisition2) == 1

    acquisition_couple = (acquisition_id1,acquisition_id2)

    voxel_size1 = acquisition1.iloc[0].at['voxel_size']
    voxel_size2 = acquisition2.iloc[0].at['voxel_size']
    shape1 = acquisition1.iloc[0].at['reordered_shape']
    shape2 = acquisition2.iloc[0].at['reordered_shape']

    if voxel_size1 != voxel_size2 : 
        raise MissMatchError("voxel size 1 different than voxel size 2")
    else :
        voxel_size = voxel_size1

    if shape1 != shape2 :
        raise MissMatchError("shape 1 different than shape 2")
    else :
        shape = shape1


    spots1 = acquisition1.iloc[0].at['spots']
    spots2 = acquisition2.iloc[0].at['spots']

    spot1_total = len(spots1)
    spot2_total = len(spots2)

    try :
        fraction_spots1_coloc_spots2 = spots_colocalisation(spot_list1=spots1, spot_list2=spots2, distance= colocalisation_distance, voxel_size=voxel_size) / spot1_total
        fraction_spots2_coloc_spots1 = spots_colocalisation(spot_list1=spots2, spot_list2=spots1, distance= colocalisation_distance, voxel_size=voxel_size) / spot2_total
    except MissMatchError as e :
        sg.popup(str(e))
        fraction_spots1_coloc_spots2 = np.NaN
        fraction_spots2_coloc_spots1 = np.NaN

    if 'clusters' in acquisition1.columns :
        try : 
            clusters_id_1 = acquisition1.iloc[0].at['spots_cluster_id']
            fraction_spots2_coloc_cluster1 = spots_colocalisation(spot_list1=spots2, spot_list2=spots1[clusters_id_1 != -1], distance= colocalisation_distance, voxel_size=voxel_size) / spot2_total
        except MissMatchError as e :
            sg.popup(str(e))
            fraction_spots2_coloc_cluster1 = np.NaN
        except TypeError : # clusters not computed
            fraction_spots2_coloc_cluster1 = np.NaN


    else : fraction_spots2_coloc_cluster1 = np.NaN

    if 'clusters' in acquisition2.columns :
        try :
            clusters_id_2 = acquisition2.iloc[0].at['spots_cluster_id']
            fraction_spots1_coloc_cluster2 = spots_colocalisation(spot_list1=spots1, spot_list2=spots2[clusters_id_2 != -1], distance= colocalisation_distance, voxel_size=voxel_size) / spot1_total
        except MissMatchError as e :# clusters not computed
            sg.popup(str(e))
            fraction_spots1_coloc_cluster2 = np.NaN
        except TypeError :
            fraction_spots1_coloc_cluster2 = np.NaN


    else : fraction_spots1_coloc_cluster2 = np.NaN

    if 'clusters' in acquisition2.columns and 'clusters' in acquisition1.columns :
        try :
            total_clustered_spots1 = len(spots1[clusters_id_1 != -1])
            total_clustered_spots2 = len(spots2[clusters_id_2 != -1])
            fraction_cluster1_coloc_cluster2 = spots_colocalisation(spot_list1=spots1[clusters_id_1 != -1], spot_list2=spots2[clusters_id_2 != -1], distance= colocalisation_distance, voxel_size=voxel_size) / total_clustered_spots1
            fraction_cluster2_coloc_cluster1 = spots_colocalisation(spot_list1=spots2[clusters_id_2 != -1], spot_list2=spots1[clusters_id_1 != -1], distance= colocalisation_distance, voxel_size=voxel_size) / total_clustered_spots2
        except MissMatchError as e :# clusters not computed
            sg.popup(str(e))
            fraction_cluster1_coloc_cluster2 = np.NaN
            fraction_cluster2_coloc_cluster1 = np.NaN
        except TypeError :
            fraction_cluster1_coloc_cluster2 = np.NaN
            fraction_cluster2_coloc_cluster1 = np.NaN

    else :
        fraction_cluster1_coloc_cluster2 = np.NaN
        fraction_cluster2_coloc_cluster1 = np.NaN
        

    coloc_df = pd.DataFrame({
        "acquisition_couple" : [acquisition_couple],
        "acquisition_id_1" : [acquisition_couple[0]],
        "acquisition_id_2" : [acquisition_couple[1]],
        "colocalisation_distance" : [colocalisation_distance],
        "spot1_total" : [spot1_total],
        "spot2_total" : [spot2_total],
        'fraction_spots1_coloc_spots2' : [fraction_spots1_coloc_spots2],
        'fraction_spots2_coloc_spots1' : [fraction_spots2_coloc_spots1],
        'fraction_spots2_coloc_cluster1' : [fraction_spots2_coloc_cluster1],
        'fraction_spots1_coloc_cluster2' : [fraction_spots1_coloc_cluster2],
        'fraction_cluster1_coloc_cluster2' : [fraction_cluster1_coloc_cluster2],
        'fraction_cluster2_coloc_cluster1' : [fraction_cluster2_coloc_cluster1],
    })

    coloc_df['fraction_spots1_coloc_free2'] = coloc_df['fraction_spots1_coloc_spots2'] - coloc_df['fraction_spots1_coloc_cluster2']
    coloc_df['fraction_spots2_coloc_free1'] = coloc_df['fraction_spots2_coloc_spots1'] - coloc_df['fraction_spots2_coloc_cluster1']

    #Add names
    coloc_df_col = list(coloc_df.columns)
    coloc_df['name1'] = acquisition1.iloc[0].at['name']
    coloc_df['name2'] = acquisition2.iloc[0].at['name']
    coloc_df = coloc_df.loc[:,['name1','name2'] + coloc_df_col]

    return coloc_df

def _cell_coloc(
        acquisition_id1: int,
        acquisition_id2: int,
        result_dataframe : pd.DataFrame, 
        cell_dataframe : pd.DataFrame, 
        colocalisation_distance : float,
        ) :
    
    acquisition1 = result_dataframe.loc[result_dataframe['acquisition_id'] == acquisition_id1]
    acquisition2 = result_dataframe.loc[result_dataframe['acquisition_id'] == acquisition_id2]

    acquisition_name_id1 = acquisition1['name'].iat[0]
    acquisition_name_id2 = acquisition2['name'].iat[0]
    result_dataframe = result_dataframe.set_index('acquisition_id', drop=False)
    coloc_name_forward = '{0} -> {1}'.format(acquisition_name_id1, acquisition_name_id2)
    coloc_name_backward = '{1} -> {0}'.format(acquisition_name_id1, acquisition_name_id2)

    #Getting shape
    if not result_dataframe.at[acquisition_id1, 'reordered_shape'] == result_dataframe.at[acquisition_id2, 'reordered_shape'] :
        raise ValueError("Selected acquisitions have different shapes. Most likely they don't belong to the same fov.")

    #Getting voxel_size
    if not result_dataframe.at[acquisition_id1, 'voxel_size'] == result_dataframe.at[acquisition_id2, 'voxel_size'] :
        raise ValueError("Selected acquisitions have different voxel_size. Most likely they don't belong to the same fov.")
    voxel_size = result_dataframe.at[acquisition_id1, 'voxel_size']

    #Selecting relevant cells in Cell table
    cell_dataframe = cell_dataframe.loc[(cell_dataframe['acquisition_id'] == acquisition_id1)|(cell_dataframe['acquisition_id'] == acquisition_id2)]

    #Putting spots lists in 2 cols for corresponding cells
    pivot_values_columns = ['rna_coords', 'total_rna_number']
    if 'clusters' in acquisition2.columns or 'clusters' in acquisition1.columns :
        pivot_values_columns.extend(['clustered_spots_coords','clustered_spot_number'])
    cell_dataframe['cell_id'] = cell_dataframe['cell_id'].astype(int)
    colocalisation_df = cell_dataframe.pivot(
        columns=['name', 'acquisition_id'],
        values= pivot_values_columns,
        index= 'cell_id'
    )
    #spots _vs spots
    colocalisation_df[("spots_with_spots_count",coloc_name_forward,"forward")] = colocalisation_df['rna_coords'].apply(
        lambda x: spots_colocalisation(
            spot_list1= x[(acquisition_name_id1,acquisition_id1)],
            spot_list2= x[(acquisition_name_id2,acquisition_id2)],
            distance=colocalisation_distance,
            voxel_size=voxel_size
            ),axis=1
        )
    colocalisation_df[("spots_with_spots_fraction",coloc_name_forward,"forward")] = colocalisation_df[("spots_with_spots_count",coloc_name_forward,"forward")].astype(float) / colocalisation_df[('total_rna_number',acquisition_name_id1,acquisition_id1)].astype(float)
    
    colocalisation_df[("spots_with_spots_count",coloc_name_backward,"backward")] = colocalisation_df['rna_coords'].apply(
        lambda x: spots_colocalisation(
            spot_list1= x[(acquisition_name_id2,acquisition_id2)],
            spot_list2= x[(acquisition_name_id1,acquisition_id1)],
            distance=colocalisation_distance,
            voxel_size=voxel_size
            ),axis=1
        )
    colocalisation_df[("spots_with_spots_fraction",coloc_name_backward,"backward")] = colocalisation_df[("spots_with_spots_count",coloc_name_backward,"backward")].astype(float) / colocalisation_df[('total_rna_number',acquisition_name_id2,acquisition_id2)].astype(float)

    if acquisition2['do_cluster_computation'].iat[0] :
        if len(acquisition2['clusters'].iat[0]) > 0 :

            #spots to clusters
            colocalisation_df[("spots_with_clustered_spots_count",coloc_name_forward,"forward")] = colocalisation_df.apply(
                lambda x: spots_colocalisation(
                    spot_list1= x[('rna_coords',acquisition_name_id1,acquisition_id1)],
                    spot_list2= x[('clustered_spots_coords',acquisition_name_id2,acquisition_id2)][:,:len(voxel_size)],
                    distance=colocalisation_distance,
                    voxel_size=voxel_size
                    ),axis=1
                )
            colocalisation_df[("spots_with_clustered_spots_fraction",coloc_name_forward,"forward")] = colocalisation_df[("spots_with_clustered_spots_count",coloc_name_forward,"forward")].astype(float) / colocalisation_df[('total_rna_number',acquisition_name_id1,acquisition_id1)].astype(float)
        
    if acquisition1['do_cluster_computation'].iat[0] :
        if len(acquisition1['clusters'].iat[0]) > 0 :
            colocalisation_df[("spots_with_clustered_spots_count",coloc_name_backward,"backward")] = colocalisation_df.apply(
                lambda x: spots_colocalisation(
                    spot_list1= x[('rna_coords',acquisition_name_id2,acquisition_id2)],
                    spot_list2= x[('clustered_spots_coords',acquisition_name_id1,acquisition_id1)][:,:len(voxel_size)],
                    distance=colocalisation_distance,
                    voxel_size=voxel_size
                    ),axis=1
                )

            colocalisation_df[("spots_with_clustered_spots_fraction",coloc_name_backward,"backward")] = colocalisation_df[("spots_with_clustered_spots_count",coloc_name_backward,"backward")].astype(float) / colocalisation_df[('total_rna_number',acquisition_name_id2,acquisition_id2)].astype(float)

    if acquisition2['do_cluster_computation'].iat[0] and acquisition1['do_cluster_computation'].iat[0] :
        if len(acquisition1['clusters'].iat[0]) > 0 and len(acquisition2['clusters'].iat[0]) > 0 :
            #clusters to clusters 
            colocalisation_df[("clustered_spots_with_clustered_spots_count",coloc_name_forward,"forward")] = colocalisation_df.apply(
                lambda x: spots_colocalisation(
                    spot_list1= x[('clustered_spots_coords',acquisition_name_id1,acquisition_id1)][:,:len(voxel_size)],
                    spot_list2= x[('clustered_spots_coords',acquisition_name_id2,acquisition_id2)][:,:len(voxel_size)],
                    distance=colocalisation_distance, 
                    voxel_size=voxel_size 
                    ),axis=1
            )
            colocalisation_df[("clustered_spots_with_clustered_spots_fraction",coloc_name_forward,"forward")] = colocalisation_df[("clustered_spots_with_clustered_spots_count",coloc_name_forward,"forward")].astype(float) / colocalisation_df[('clustered_spot_number',acquisition_name_id1,acquisition_id1)].astype(float)

            colocalisation_df[("clustered_spots_with_clustered_spots_count",coloc_name_backward,"backward")] = colocalisation_df.apply(
                lambda x: spots_colocalisation(
                    spot_list1= x[('clustered_spots_coords',acquisition_name_id2,acquisition_id2)][:,:len(voxel_size)],
                    spot_list2= x[('clustered_spots_coords',acquisition_name_id1,acquisition_id1)][:,:len(voxel_size)],
                    distance=colocalisation_distance, 
                    voxel_size=voxel_size 
                    ),axis=1
            )
            colocalisation_df[("clustered_spots_with_clustered_spots_fraction",coloc_name_backward,"backward")] = colocalisation_df[("clustered_spots_with_clustered_spots_count",coloc_name_backward,"backward")].astype(float) / colocalisation_df[('clustered_spot_number',acquisition_name_id2,acquisition_id2)].astype(float)

    colocalisation_df = colocalisation_df.sort_index(axis=0).sort_index(axis=1, level=0)

    if 'clustered_spots_coords' in cell_dataframe.columns : colocalisation_df = colocalisation_df.drop('clustered_spots_coords', axis=1)
    colocalisation_df = colocalisation_df.drop('rna_coords', axis=1)
    colocalisation_df['voxel_size'] = [voxel_size]*len(colocalisation_df)
    colocalisation_df['pair_name'] = [(acquisition_name_id1, acquisition_name_id2)] * len(colocalisation_df)
    colocalisation_df['acquisition_id_1'] = [acquisition_id1] * len(colocalisation_df)
    colocalisation_df['acquisition_id_2'] = [acquisition_id2] * len(colocalisation_df)
    colocalisation_df['colocalisation_distance'] = colocalisation_distance

    return colocalisation_df

@add_default_loading
def launch_colocalisation(result_tables, result_dataframe, cell_result_dataframe, colocalisation_distance, global_coloc_df, cell_coloc_df: dict) :
    
    acquisition1 = result_dataframe.iloc[result_tables[0]]
    acquisition2 = result_dataframe.iloc[result_tables[1]]

    acquisition_id1, acquisition_id2 = (acquisition1.at['acquisition_id'], acquisition2.at['acquisition_id'])

    if acquisition_id1 in list(cell_result_dataframe['acquisition_id']) and acquisition_id2 in list(cell_result_dataframe['acquisition_id']) :
        print("Launching cell to cell colocalisation.")
        new_coloc = _cell_coloc(
            acquisition_id1 = acquisition_id1,
            acquisition_id2 = acquisition_id2,
            result_dataframe = result_dataframe,
            cell_dataframe=cell_result_dataframe,
            colocalisation_distance=colocalisation_distance
        )
        
        index = 0
        while (acquisition_id1, acquisition_id2, index) in cell_coloc_df.keys() :
            index +=1
        cell_coloc_df [(acquisition_id1,acquisition_id2, index)] = new_coloc


    else :
        print("Launching global colocalisation.")
        new_coloc = _global_coloc(
            acquisition_id1=acquisition_id1,
            acquisition_id2=acquisition_id2,
            result_dataframe=result_dataframe,
            colocalisation_distance=colocalisation_distance,
        )
        global_coloc_df = pd.concat([
            global_coloc_df,
            new_coloc,
        ], axis=0).reset_index(drop=True)


    return global_coloc_df, cell_coloc_df


"""
This subpackge contains code related to graphical interface
"""

from .prompts import _error_popup
from .prompts import _warning_popup 
from .prompts import prompt
from .prompts import input_image_prompt
from .prompts import hub_prompt
from .prompts import detection_parameters_promt
from .prompts import coloc_prompt
from .prompts import output_image_prompt
from .prompts import ask_cancel_detection
from .prompts import ask_cancel_segmentation
from .prompts import ask_detection_confirmation
from .prompts import prompt_restore_main_menu

#Helpers to build windows
from .layout import parameters_layout 
from .layout import bool_layout
from .layout import path_layout
from .layout import combo_elmt
from .layout import tuple_layout
from .layout import radio_layout
from .layout import add_header

from .animation import add_default_loading


import FreeSimpleGUI as sg
import os
from ..utils import check_parameter
from ..hints import pipeline_parameters
from typing import Optional, Union
import cellpose.models as models
from cellpose.core import use_gpu

sg.theme('DarkAmber')


def add_header(header_text) :
    """Returns [elmnt] not layout"""
    header = [sg.Text('\n{0}'.format(header_text), size= (len(header_text),3), font= 'bold 15')]
    return header


def pad_right(string, length, pad_char) :
    if len(string) >= length : return string
    else : return string + pad_char* (length - len(string))
    

def parameters_layout(parameters:'list[str]' = [], unit=None, header= None, default_values=None, size=5, opt:list=None) :

    if len(parameters) == 0 : return []
    check_parameter(parameters= list, header = (str, type(None)))
    for key in parameters : check_parameter(key = str)
    max_length = len(max(parameters, key=len))

    if type(opt) == type(None) :
        opt = [False] * len(parameters)
    else :
        if len(opt) != len(parameters) : raise ValueError("Parameters and opt must be of same length.")

    if isinstance(default_values, (list, tuple)) :
        if len(default_values) != len(parameters) : raise ValueError("if default values specified it must be of equal length as parameters.")
        layout= [
            [sg.Text("{0}".format(pad_right(parameter, max_length, ' ')), text_color= 'green' if option else None), 
             sg.InputText(size= size, key= parameter, default_text= value)
             
             ] for parameter,value, option in zip(parameters,default_values, opt)
        ]
    else :
        layout= [
            [sg.Text("{0}".format(pad_right(parameter, max_length, ' ')), text_color= 'green' if option else None), 
             sg.InputText(size= size, key= parameter)
             
             ] for parameter, option in zip(parameters, opt)
        ]
    
    if type(unit) == str :
        for line_id, line in enumerate(layout) :
            layout[line_id] += [sg.Text('{0}'.format(unit))]
    
    if isinstance(header, str) :
        layout = [add_header(header)] + layout
    return layout

def tuple_layout(opt=None, default_dict={}, unit:dict={}, **tuples) :
    """
    tuples example : voxel_size = ['z','y','x']; will ask a tuple with 3 element default to 'z', 'y' 'x'.
    """
    if type(tuples) == type(None) : return []
    if len(tuples.keys()) == 0 : return []

    if type(opt) != type(None) :
        if not isinstance(opt, dict) : raise TypeError("opt parameter should be either None or dict type.")
        if not opt.keys() == tuples.keys() : raise ValueError("If opt is passed it is expected to have same keys as tuples dict.")
    else : 
        opt = tuples.copy()
        for key in opt.keys() :
            opt[key] = False

    for tup in tuples : 
        if not isinstance(tuples[tup], (list,tuple)) : raise TypeError()

    max_size = len(max(tuples.keys(), key=len))
    
    layout = [
        [sg.Text(pad_right(tup, max_size, ' '), text_color= 'green' if opt[option] else None)] 
        + [sg.InputText(default_text=default_dict.setdefault('{0}_{1}'.format(tup,elmnt), elmnt),key= '{0}_{1}'.format(tup, elmnt), size= 5) for elmnt in tuples[tup]]
        + [sg.Text(unit.setdefault(tup,''))] 
        for tup,option, in zip(tuples,opt)
    ]

    return layout

def path_layout(keys= [],look_for_dir = False, header=None, preset=os.getcwd()) :
    """
    If not look for dir then looks for file.
    """
    if len(keys) == 0 : return []
    check_parameter(keys= list, header = (str, type(None)))
    for key in keys : check_parameter(key = str)
    if look_for_dir : Browse = sg.FolderBrowse
    else : Browse = sg.FileBrowse

    max_length = len(max(keys, key=len))
    layout = [
        [sg.Text(pad_right(name, max_length, ' ')), Browse(key= name, target=(555666777,2), initial_folder= preset), sg.Text('')] for name in keys
        ]
    if isinstance(header, str) :
        layout = [add_header(header)] + layout
    return layout

def bool_layout(parameters= [], header=None, preset : Optional[Union['list[bool]',bool,None]]=None, keys=None) :
    if len(parameters) == 0 : return []
    check_parameter(parameters= list, header= (str, type(None)), preset=(type(None), list, tuple, bool))
    for key in parameters : check_parameter(key = str)
    if type(preset) == type(None) :
        preset = [False] * len(parameters)
    elif type(preset) == bool :
        preset = [preset] * len(parameters)
    else : 
        for key in preset : check_parameter(key = bool)

    max_length = len(max(parameters, key=len))

    if type(keys) == type(None) :
        keys = parameters
    elif isinstance(keys,(list,tuple)) :
        if len(keys) != len(parameters) : raise ValueError("keys arguement must be of same length than parameters argument")
    elif len(parameters) == 1 and isinstance(keys, str) :
        keys = [keys]
    else :
        raise ValueError('Incorrect keys parameters. Expected list of same length than parameters or None.')

    layout = [
        [sg.Checkbox(pad_right(name, max_length, ' '), key=key, default=box_preset)] for name, box_preset, key in zip(parameters,preset,keys)
    ]
    if isinstance(header, str) :
        layout = [add_header(header)] + layout
    return layout

def combo_elmt(values, key, header=None, read_only=True, default_value=None) :
    """
    drop-down list
    """
    if len(values) == 0 : return []
    check_parameter(values= list, header= (str, type(None)))
    if type(default_value) == type(None) :
        default_value = values[0]
    elif default_value not in values :
        default_value = values[0]
    layout = [
        sg.Combo(values, default_value=default_value, readonly=read_only, key=key)
    ]
    if isinstance(header, str) :
        layout = add_header(header) + layout
    return layout

def radio_layout(values, header=None, key=None) :
    """
    Single choice buttons.
    """
    if len(values) == 0 : return []
    check_parameter(values= list, header= (str, type(None)))
    layout = [
        [sg.Radio(value, group_id= 0, key=key) for value in values]
    ]
    if isinstance(header, str) :
        layout = [add_header(header)] + layout
    return layout

def _segmentation_layout(
        multichannel, 
        cytoplasm_model_preset= 'cyto3', 
        nucleus_model_preset= 'nuclei', 
        cytoplasm_channel_preset=0, 
        nucleus_channel_preset=0, 
        other_nucleus_image_preset = None,
        cyto_diameter_preset=30, 
        nucleus_diameter_preset= 30, 
        show_segmentation_preset= False, 
        segment_only_nuclei_preset=False, 
        saving_path_preset=os.getcwd(), 
        filename_preset='cell_segmentation.png',
        ) :
    
    USE_GPU = use_gpu()

    models_list = models.get_user_models() + models.MODEL_NAMES
    if len(models_list) == 0 : models_list = ['no model found']
    
    #Header : GPU availabality
    layout = [
        [sg.Text("GPU is currently "), sg.Text('ON', text_color= 'green') if USE_GPU else sg.Text('OFF', text_color= 'red')]
        ]
    
    #cytoplasm parameters
    layout += [
        add_header("Cell Segmentation"),
        [sg.Text("Choose cellpose model for cytoplasm: \n")],
        combo_elmt(models_list, key='cyto_model_name', default_value= cytoplasm_model_preset)
                        ]
                        
    if multichannel : layout += parameters_layout(['cytoplasm channel'],default_values= [cytoplasm_channel_preset])


    layout += parameters_layout(['cytoplasm diameter'], unit= "px", default_values= [cyto_diameter_preset])
    #Nucleus parameters
    layout += [
            add_header("Nucleus segmentation"),
            [sg.Text("Choose cellpose model for nucleus: \n")],
              combo_elmt(models_list, key='nucleus_model_name', default_value= nucleus_model_preset)
                ]
    
    if multichannel : layout += parameters_layout(['nucleus channel'], default_values= [nucleus_channel_preset])
    layout += path_layout(['other_nucleus_image'], preset=other_nucleus_image_preset)
    layout += parameters_layout([ 'nucleus diameter'],unit= "px", default_values= [nucleus_diameter_preset])
    layout += bool_layout(["Segment only nuclei"], preset=segment_only_nuclei_preset)
    
    #Control plots
    layout += bool_layout(['show segmentation'], header= 'Segmentation plots', preset= show_segmentation_preset)
    layout += path_layout(['saving path'], look_for_dir=True, preset=saving_path_preset)
    layout += parameters_layout(['filename'], default_values=[filename_preset], size= 25)

    return layout

def _input_parameters_layout(
        ask_for_segmentation,
        is_3D_stack_preset,
        time_stack_preset,
        multichannel_preset,
        do_dense_regions_deconvolution_preset,
        do_clustering_preset,
        do_segmentation_preset,
        do_Napari_correction

) :
    layout_image_path = path_layout(['image_path'], header= "Image")
    layout_image_path += bool_layout(['3D stack', 'Multichannel stack'], keys=['is_3D_stack', 'is_multichannel'], preset= [is_3D_stack_preset, multichannel_preset])
    
    layout_image_path += bool_layout(
        ['Dense regions deconvolution', 'Compute clusters', 'Cell segmentation', 'Open Napari corrector'],
        keys= ['do_dense_regions_deconvolution', 'do_cluster_computation', 'do_segmentation', 'show_napari_corrector'], 
        preset= [do_dense_regions_deconvolution_preset, do_clustering_preset, do_segmentation_preset, do_Napari_correction], 
        header= "Pipeline settings")


    return layout_image_path

def _detection_layout(
        is_3D_stack,
        is_multichannel,
        do_dense_region_deconvolution,
        do_clustering,
        do_segmentation,
        segmentation_done=False,
        default_dict : pipeline_parameters={},
) :
    if is_3D_stack : dim = 3
    else : dim = 2

    #Detection
    detection_parameters = ['threshold', 'threshold penalty']
    default_detection = [default_dict.setdefault('threshold',''), default_dict.setdefault('threshold penalty', '1')]
    opt= [True, True]
    if is_multichannel : 
        detection_parameters += ['channel_to_compute']
        opt += [False]
        default_detection += [default_dict.setdefault('channel_to_compute', '')]
    
    layout = [[sg.Text("Green parameters", text_color= 'green'), sg.Text(" are optional parameters.")]]
    layout += parameters_layout(detection_parameters, header= 'Detection', opt=opt, default_values=default_detection)
    
    if dim == 2 : tuple_shape = ('y','x')
    else : tuple_shape = ('z','y','x')
    opt = {'voxel_size' : False, 'spot_size' : False, 'log_kernel_size' : True, 'minimum_distance' : True}
    unit = {'voxel_size' : 'nm', 'minimum_distance' : 'nm', 'spot_size' : 'radius(nm)', 'log_kernel_size' : 'px'}

    layout += tuple_layout(opt=opt, unit=unit, default_dict=default_dict, voxel_size= tuple_shape, spot_size= tuple_shape, log_kernel_size= tuple_shape, minimum_distance= tuple_shape)
    
    if (do_segmentation and is_multichannel) or (is_multichannel and segmentation_done):
        default_segmentation = [default_dict.setdefault('nucleus channel signal', default_dict.setdefault('nucleus channel',0))]
        layout += [[sg.Text("nucleus channel signal "), sg.InputText(default_text=default_segmentation, key= "nucleus channel signal", size= 5, tooltip= "Channel from which signal will be measured for nucleus features, \nallowing you to measure signal from a different channel than the one used for segmentation.")]]
        # layout += parameters_layout(['nucleus channel signal'], default_values=default_segmentation) + [[sg.Text("Channel from which signal will be measured for nucleus features, allowing you to measure signal from a different channel than the one used for segmentation.")]]
    
    #Deconvolution
    if do_dense_region_deconvolution :
        default_dense_regions_deconvolution = [default_dict.setdefault('alpha',0.5), default_dict.setdefault('beta',1)]
        layout += parameters_layout(['alpha', 'beta',], default_values= default_dense_regions_deconvolution, header= 'do_dense_regions_deconvolution')
        layout += parameters_layout(['gamma'], unit= 'px', default_values= [default_dict.setdefault('gamma',5)])
        layout += tuple_layout(opt= {"deconvolution_kernel" : True}, unit= {"deconvolution_kernel" : 'px'}, default_dict=default_dict, deconvolution_kernel = tuple_shape)
    
    #Clustering
    if do_clustering :
        layout += parameters_layout(['cluster_size'], unit="radius(nm)", default_values=[default_dict.setdefault('cluster_size',400)])
        layout += parameters_layout(['min_number_of_spots'], default_values=[default_dict.setdefault('min_number_of_spots', 5)])

    

    layout += bool_layout(['Interactive threshold selector'],keys = ['show_interactive_threshold_selector'], preset=[False])
    layout += path_layout(
        keys=['spots_extraction_folder'],
        look_for_dir=True,
        header= "Individual spot extraction",
        preset= default_dict.setdefault('spots_extraction_folder', '')
    )
    default_filename = default_dict.setdefault("filename","") + "_spot_extraction"
    layout += parameters_layout(
        parameters=['spots_filename'],
        default_values=[default_filename],
        size= 13
    )
    layout += bool_layout(
        parameters= ['do_spots_csv', 'do_spots_excel', 'do_spots_feather'],
        preset= [default_dict.setdefault('do_spots_csv',False), default_dict.setdefault('do_spots_excel',False),default_dict.setdefault('do_spots_feather',False)]
    )

    return layout

def _ask_channel_map_layout(
        shape,
        is_3D_stack,
        multichannel,
        is_time_stack,
        preset_map={},
) :
    
    x = preset_map.setdefault('x',0)
    y = preset_map.setdefault('y',0)
    z = preset_map.setdefault('z',0)
    c = preset_map.setdefault('c',0)
    t = preset_map.setdefault('t',0)

    layout = [
            add_header("Dimensions mapping"), [sg.Text("Image shape : {0}".format(shape))]
        ]
    layout += parameters_layout(['x','y'], default_values=[x,y])
    if is_3D_stack : layout += parameters_layout(['z'], default_values=[z])
    if multichannel : layout += parameters_layout(['c'], default_values=[c])
    if is_time_stack : layout += parameters_layout(['t'], default_values=[t])

    return layout


"""
This code handles the exchange between the computer and the code. That is to say opening and saving data.
"""

from .image import open_image
from .image import get_filename 
from .image import check_format
from .image import FormatError
from .image import get_voxel_size

from .inoutput import write_results


import os
import pandas as pd
import numpy as np
from bigfish.stack import check_parameter
from typing import Literal

MAX_LEN_EXCEL = 1048576 #Maximum number of lines that can be written in an excel file

def _cast_spot_to_tuple(spot) :
    return tuple([coord for coord in spot])

def _cast_spots_to_tuple(spots) :
    return tuple(list(map(_cast_spot_to_tuple, spots)))

def write_list_of_results(
        result_list : list,
        path : str,
        filename : str,
        do_excel = True,
        do_feather = False,
        do_csv = False,
        ) :
    

    # FORMAT CHECKING
    if len(result_list) == 0 : return True
    if not do_excel and not do_feather and not do_csv : 
        return False
    elif do_feather and not do_csv and not do_excel: 
        print("WARNING : cell_to_cell colocalisation : .feather is depreciated saving as csv instead.")
        do_csv = True
    elif do_feather :
        print("WARNING : cell_to_cell colocalisation : .feather is depreciated will be removed in future version.")
    
    #PATH CHECKING
    if not path.endswith('/') : path +='/'
    assert os.path.isdir(path)

    #NO OVERWRITE CHECKING
    new_filename = filename

    i= 1
    while new_filename + '.xlsx' in os.listdir(path) or new_filename + '.csv' in os.listdir(path) :
        new_filename = filename + '_{0}'.format(i)
        i+=1

    if do_excel :
        start_row = 0
        sheet_number = 1
        sheet_name = "Sheet" + str(sheet_number)
        with pd.ExcelWriter(path + new_filename + '.xlsx', engine='openpyxl') as writer:
            for df in result_list :

                if start_row + len(df) > 1e6 : #Excel limit of 1 milion line.
                    print("To many excel lines, changing excel sheet.")
                    sheet_number += 1
                    sheet_name = "Sheet" + str(sheet_number)
                    start_row = 0

                #writing
                df.to_excel(writer, sheet_name=sheet_name, startrow=start_row)
                start_row += len(df) + 5
                
    
    if do_csv :
        with open(path + new_filename + '.csv', 'w') as f:
            for df in result_list :
                df.to_csv(f)
                f.write("\n\n")  # Add 2 lines in between DataFrames.
    
    return True


def write_results(dataframe: pd.DataFrame, path:str, filename:str, do_excel= True, do_feather= False, do_csv=False, overwrite=False, reset_index=True) :
    check_parameter(dataframe= pd.DataFrame, path= str, filename = str, do_excel = bool, do_feather = bool)

    if len(dataframe) == 0 : return True
    if not do_excel and not do_feather and not do_csv : 
        return False

    if not path.endswith('/') : path +='/'
    assert os.path.isdir(path)

    #Casting cols name to str for feather format
    index_dim = dataframe.columns.nlevels
    if index_dim == 1 :
        dataframe.columns = dataframe.columns.astype(str)
    else :
        casted_cols = [dataframe.columns.get_level_values(level).astype(str) for level in range(index_dim)]
        casted_cols = zip(*casted_cols)
        dataframe.columns = pd.MultiIndex.from_tuples(casted_cols)

    new_filename = filename
    i= 1
    if not overwrite :
        while new_filename + '.xlsx' in os.listdir(path) or new_filename + '.parquet' in os.listdir(path) or new_filename + '.csv' in os.listdir(path) :
            new_filename = filename + '_{0}'.format(i)
            i+=1

    COLUMNS_TO_DROP = ['image', 'spots', 'clusters', 'rna_coords', 'cluster_coords']
    for col in COLUMNS_TO_DROP :
        if col in dataframe.columns : dataframe = dataframe.drop(columns=col)

    if reset_index : dataframe = dataframe.reset_index(drop=True)

    if do_csv : dataframe.to_csv(path + new_filename + '.csv', sep=";")
    if do_excel : 
        if len(dataframe) < MAX_LEN_EXCEL :
            dataframe.to_excel(path + new_filename + '.xlsx')
        else : 
            print("Error : Table too big to be saved in excel format.")
            return False
    
    if do_feather : 
        dataframe.to_parquet(path + new_filename + '.parquet')

    return True

def input_segmentation(
        nucleus_path : str,
        cytoplasm_path : str,
) :
    
    if nucleus_path.endswith('.npy') or nucleus_path.endswith('.npz') :
        nucleus_label = np.load(nucleus_path)
    else :
        raise ValueError("Wrong extension for mask file, only npy and npz are supported")


    if cytoplasm_path != '' :

        if cytoplasm_path.endswith('.npy') or cytoplasm_path.endswith('.npz') :
            cytoplasm_label = np.load(cytoplasm_path)
        else :
            raise ValueError("Wrong extension for mask file, only npy and npz are supported")


    else : 
        cytoplasm_label = nucleus_label

    return nucleus_label, cytoplasm_label

def output_segmentation(
        path : str,
        extension : Literal['npy', 'npz_uncompressed', 'npz_compressed'],
        nucleus_label : np.ndarray,
        cytoplasm_label : np.ndarray = None,
) :
    
    saved = False
    if extension == 'npy' :
        save = np.save
    elif extension == 'npz_uncompressed' :
        save = np.savez
    elif extension == 'npz_compressed' :
        save = np.savez_compressed

    if type(nucleus_label) != type(None) :
        save(path + "_nucleus_segmentation", nucleus_label)
        saved = True
    
    if type(cytoplasm_label) != type(None) :
        save(path + "_cytoplasm_segmentation", cytoplasm_label)

    return saved


"""
Signal to noise wrapper from BigFish code.
"""


# ### SNR ###
import bigfish.stack as stack
import numpy as np
from bigfish.detection.utils import get_object_radius_pixel, get_spot_volume, get_spot_surface

def compute_snr_spots(image, spots, voxel_size, spot_radius):
    """
    Modified version of bigfish.detection.utils compute_snr_spots : 
    # Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
    # License: BSD 3 clause
    
    Compute signal-to-noise ratio (SNR) based on spot coordinates.

    .. math::

        \\mbox{SNR} = \\frac{\\mbox{max(spot signal)} -
        \\mbox{mean(background)}}{\\mbox{std(background)}}

    Background is a region twice larger surrounding the spot region. Only the
    y and x dimensions are taking into account to compute the SNR.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    spots : np.ndarray
        Coordinate of the spots, with shape (nb_spots, 3) or (nb_spots, 2).
        One coordinate per dimension (zyx or yx coordinates).
    voxel_size : int, float, Tuple(int, float), List(int, float) or None
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions. Not used if 'log_kernel_size' and 'minimum_distance' are
        provided.
    spot_radius : int, float, Tuple(int, float), List(int, float) or None
        Radius of the spot, in nanometer. One value per spatial dimension (zyx
        or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions. Not used if 'log_kernel_size' and 'minimum_distance'
        are provided.

    Returns
    -------
    res : dict
        Median signal-to-noise ratio computed for every spots.
        +
            'snr_median' : np.median(snr_spots),
            'snr_mean' : np.mean(snr_spots),
            'snr_std' : np.std(snr_spots),
            'cell_medianbackground_mean' : np.mean(median_background_list),
            'cell_medianbackground_std' : np.std(median_background_list),
            'cell_meanbackground_mean'  : np.mean(mean_background_list),
            'cell_meanbackground_std'  : np.std(mean_background_list),
            'cell_stdbackground_mean' : np.mean(std_background_list),
            'cell_stdbackground_std' : np.std(std_background_list)

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_range_value(image, min_=0)
    stack.check_array(
        spots,
        ndim=2,
        dtype=[np.float32, np.float64, np.int32, np.int64])
    stack.check_parameter(
        voxel_size=(int, float, tuple, list),
        spot_radius=(int, float, tuple, list))

    # check consistency between parameters
    ndim = image.ndim
    if ndim != spots.shape[1]:
        raise ValueError("Provided image has {0} dimensions but spots are "
                         "detected in {1} dimensions."
                         .format(ndim, spots.shape[1]))
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError(
                "'voxel_size' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim
    if isinstance(spot_radius, (tuple, list)):
        if len(spot_radius) != ndim:
            raise ValueError(
                "'spot_radius' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        spot_radius = (spot_radius,) * ndim

    # cast spots coordinates if needed
    if spots.dtype == np.float64:
        spots = np.round(spots).astype(np.int64)

    # cast image if needed
    image_to_process = image.copy().astype(np.float64)

    # clip coordinate if needed
    if ndim == 3:
        spots[:, 0] = np.clip(spots[:, 0], 0, image_to_process.shape[0] - 1)
        spots[:, 1] = np.clip(spots[:, 1], 0, image_to_process.shape[1] - 1)
        spots[:, 2] = np.clip(spots[:, 2], 0, image_to_process.shape[2] - 1)
    else:
        spots[:, 0] = np.clip(spots[:, 0], 0, image_to_process.shape[0] - 1)
        spots[:, 1] = np.clip(spots[:, 1], 0, image_to_process.shape[1] - 1)

    # compute radius used to crop spot image
    radius_pixel = get_object_radius_pixel(
        voxel_size_nm=voxel_size,
        object_radius_nm=spot_radius,
        ndim=ndim)
    radius_signal_ = [np.sqrt(ndim) * r for r in radius_pixel]
    radius_signal_ = tuple(radius_signal_)

    # compute the neighbourhood radius
    radius_background_ = tuple(i * 2 for i in radius_signal_)

    # ceil radii
    radius_signal = np.ceil(radius_signal_).astype(int)
    radius_background = np.ceil(radius_background_).astype(int)

    # loop over spots
    snr_spots = []
    median_background_list = []
    mean_background_list = []
    std_background_list = []
    

    for spot in spots:

        # extract spot images
        spot_y = spot[ndim - 2]
        spot_x = spot[ndim - 1]
        radius_signal_yx = radius_signal[-1]
        radius_background_yx = radius_background[-1]
        edge_background_yx = radius_background_yx - radius_signal_yx
        if ndim == 3:
            spot_z = spot[0]
            radius_background_z = radius_background[0]
            max_signal = image_to_process[spot_z, spot_y, spot_x]
            spot_background_, _ = get_spot_volume(
                image_to_process, spot_z, spot_y, spot_x,
                radius_background_z, radius_background_yx)
            spot_background = spot_background_.copy()

            # discard spot if cropped at the border (along y and x dimensions)
            expected_size = (2 * radius_background_yx + 1) ** 2
            actual_size = spot_background.shape[1] * spot_background.shape[2]
            if expected_size != actual_size:
                continue

            # remove signal from background crop
            spot_background[:,
                            edge_background_yx:-edge_background_yx,
                            edge_background_yx:-edge_background_yx] = -1
            spot_background = spot_background[spot_background >= 0]

        else:
            max_signal = image_to_process[spot_y, spot_x]
            spot_background_, _ = get_spot_surface(
                image_to_process, spot_y, spot_x, radius_background_yx)
            spot_background = spot_background_.copy()

            # discard spot if cropped at the border
            expected_size = (2 * radius_background_yx + 1) ** 2
            if expected_size != spot_background.size:
                continue

            # remove signal from background crop
            spot_background[edge_background_yx:-edge_background_yx,
                            edge_background_yx:-edge_background_yx] = -1
            spot_background = spot_background[spot_background >= 0]

        # compute mean background
        median_background = np.median(spot_background)
        mean_background = np.mean(spot_background)

        # compute standard deviation background
        std_background = np.std(spot_background)

        # compute SNR
        snr = (max_signal - mean_background) / std_background
        snr_spots.append(snr)
        median_background_list.append(median_background)
        mean_background_list.append(mean_background)
        std_background_list.append(std_background)

    #  average SNR
    if len(snr_spots) == 0:
        res = {
            'snr_median' : 0,
            'snr_mean' : 0,
            'snr_std' : 0,
            'cell_medianbackground_mean' : 0,
            'cell_medianbackground_std' : 0,
            'cell_meanbackground_mean'  : 0,
            'cell_meanbackground_std'  : 0,
            'cell_stdbackground_mean' : 0,
            'cell_stdbackground_std' : 0
        }
    else:
        res = {
            'snr_median' : np.median(snr_spots),
            'snr_mean' : np.mean(snr_spots),
            'snr_std' : np.std(snr_spots),
            'cell_medianbackground_mean' : np.mean(median_background_list),
            'cell_medianbackground_std' : np.std(median_background_list),
            'cell_meanbackground_mean'  : np.mean(mean_background_list),
            'cell_meanbackground_std'  : np.std(mean_background_list),
            'cell_stdbackground_mean' : np.mean(std_background_list),
            'cell_stdbackground_std' : np.std(std_background_list)
        }

    return res


"""
Submodule handling handling files and filenames in batch mode.
"""

import os
import bigfish.stack as stack
import czifile as czi
import numpy as np
from .integrity import check_file

def open_image(filename:str) :

    if filename.endswith('.czi') :
        image = czi.imread(filename)
    else :
        image = stack.read_image(filename)

    image = np.squeeze(image)

    return image

def get_images(filename:str) :
    """returns filename if is image else return None"""

    supported_types = ('.tiff', '.tif', '.png', '.czi')
    if filename.endswith(supported_types) : 
        return [filename]
    else :
        return None

def get_files(path) :

    filelist = os.listdir(path)
    filelist = list(map(get_images,filelist))

    while None in filelist : filelist.remove(None)

    return filelist

def extract_files(filenames: list) :
    return sum(filenames,[])

def load(
        batch_folder:str,
) :
    if not os.path.isdir(batch_folder) :
        print("Can't open {0}".format(batch_folder))
        files_values = [[]]
        last_shape = None
        dim_number = 0
    else :
        files_values = get_files(batch_folder)
        if len(files_values) == 0 :
            last_shape = None
            dim_number = 0
        else :
            first_filename = files_values[0][0]
            last_shape = check_file(batch_folder + '/' + first_filename)
            dim_number = len(last_shape)

    
    return files_values, last_shape, dim_number


def yielder(n) :
    for i in range(n) : 
        yield i

def constructor(n) :
    def inner() :
        return yielder(n)
    return inner

#test

N = 10

gen1 = yielder(N)

a = constructor(10)

print(type(a))

gen2 = a()
gen3 = a()

print('gen1 : ', list(gen1))
print('gen2 : ', list(gen2))
print('gen3 : ', list(gen3))
print('gen3 : ', list(gen3))


#Napari test

import napari
import numpy as np
from napari.utils.events import Event
from napari.layers import Points

seed = 0
points_number = 500
rand_gen = np.random.default_rng(seed=seed)

im_shape = (10,1000,1000)
scale = (1,1,1)

im = np.zeros(shape=im_shape)

points = np.array(
    list(zip(
        rand_gen.integers(0,im_shape[0],size=points_number), #Z
        rand_gen.integers(0,im_shape[1],size=points_number), #Y
        rand_gen.integers(0,im_shape[2],size=points_number), #X
        )),
    dtype= int
)

spots_number = rand_gen.integers(0,20,points_number, dtype=int)
id_list = np.arange(points_number)

Viewer = napari.Viewer(title= "test", show=False)
Viewer.add_image(im),
cluster_layer = Viewer.add_points(
    points, 
    size = 10, 
    scale=scale, 
    face_color= 'blue', 
    opacity= 0.7, 
    symbol= 'diamond', 
    name= 'foci', 
    features= {"spot_number" : spots_number, "id" : id_list}, 
    feature_defaults= {"spot_number" : 0, "id" : -1})


class Points_callback :
    """
    Custom class to handle points number evolution during Napari run.
    """
    
    def __init__(self, points, next_id) -> None:
        self.points = points
        self.next_id = next_id
        self._set_callback()
    
    def __str__(self) -> str:
        string = 'Points_callback object state :\ncurrent_points_number : {0}\ncurrnet_id : {1}'.format(self.current_points_number, self.next_id)
        return string
    
    def get_points(self) :
        return self.points
    
    def get_next_id(self) : 
        return self.next_id
    
    def _set_callback(self) :
        def callback(event:Event) :

            old_points = self.get_points()
            new_points:Points = event.source.data
            features = event.source.features
            
            current_point_number = len(old_points)
            next_id = self.get_next_id()
            new_points_number = len(new_points)

            if new_points_number > current_point_number :
                features.at[new_points_number - 1, "id"] = next_id
                self.next_id += 1

            #preparing next callback
            self.points = new_points
            self._set_callback()
        self.callback = callback

live_points_number = len(points)
next_point_id = id_list[-1] + 1
_callback = Points_callback(points=points, next_id=next_point_id)
Viewer.show(block=False)
points_callback = Viewer.layers[1].events.data.connect((_callback, 'callback'))
napari.run()



import FreeSimpleGUI as sg
from ..pipeline._preprocess import _auto_map_channels
from ..pipeline._preprocess import MappingError

def get_elmt_from_key(Tab_elmt:sg.Tab, key) -> sg.Element:
    elmt_list = sum(Tab_elmt.Rows,[])
    for elmt in elmt_list :
        if elmt.Key == key : return elmt
    raise KeyError("{0} key not found amongst {1}.".format(key, [elmt.Key for elmt in elmt_list]))

def call_auto_map(
        tab_elmt: sg.Tab,
        shape,
        is_3D,
        is_multichannel,
    ) :
    
    if len(shape) < 2 + is_3D + is_multichannel :
        sg.popup("Image is of dimension {0} and you're trying to map_ {1} dimensions".format(len(shape), 2+is_3D+is_multichannel))
        return {}

    #Get auto map_
    try :
        map_ = _auto_map_channels(
            is_3D_stack=is_3D,
            is_time_stack=False,
            multichannel=is_multichannel,
            image=None,
            shape=shape
            )
    
    except MappingError as e :
        sg.popup_error(e)
        return {}
    
    else :
        #Acess elemnt
        x_elmt = get_elmt_from_key(tab_elmt, 'x')
        y_elmt = get_elmt_from_key(tab_elmt, 'y')
        z_elmt = get_elmt_from_key(tab_elmt, 'z')
        c_elmt = get_elmt_from_key(tab_elmt, 'c')

        #Update values
        x_elmt.update(value=map_.get('x'))
        y_elmt.update(value=map_.get('y'))
        z_elmt.update(value=map_.get('z'))
        c_elmt.update(value=map_.get('c'))

        return map_

def create_map(
        values:dict,
        is_3D:bool,
        is_multichannel:bool
        ) :
    
    maping ={
        'x': values.get('x'),
        'y': values.get('y')
    }

    if is_3D : maping['z'] = values.get('z')
    if is_multichannel : maping['c'] = values.get('c')
    
    return maping




import numpy as np

from math import ceil
from itertools import zip_longest
from skimage.measure import regionprops_table


def from_label_get_centeroidscoords(label: np.ndarray):
    """
    Returns
    --------
      centroid : dict{"label": list, "centroid-n": list} 
        n should be replace with 1,2.. according to the axe you wish to access."""

    centroid = regionprops_table(label, properties= ["label","centroid"])
    return centroid


"""
Contains cellpose wrappers to segmentate images.
"""

from cellpose.core import use_gpu
from skimage.measure import label
from ..hints import pipeline_parameters
from ..gui.layout import _segmentation_layout
from ..gui import prompt, ask_cancel_segmentation
from ..interface import open_image
from ..gui.napari_visualiser import show_segmentation as napari_show_segmentation
from .utils import from_label_get_centeroidscoords
from ._preprocess import ask_input_parameters
from ._preprocess import map_channels, reorder_shape, reorder_image_stack
from matplotlib.colors import ListedColormap

import matplotlib as mpl
import cellpose.models as models
import numpy as np
import bigfish.multistack as multistack
import bigfish.stack as stack
import bigfish.plot as plot
import FreeSimpleGUI as sg
import matplotlib.pyplot as plt
import os

def launch_segmentation(user_parameters: pipeline_parameters, nucleus_label, cytoplasm_label) :
    """
    Ask user for necessary parameters and perform cell segmentation (cytoplasm + nucleus) with cellpose.

    Input
    -----
    Image : np.ndarray[c,z,y,x]
        Image to use for segmentation.

    Returns
    -------
        cytoplasm_label, nucleus_label, user_parameters
    """

    segmentation_parameters = user_parameters.copy()

    #Ask for image parameters
    new_parameters = ask_input_parameters(ask_for_segmentation= True) #The image is open and stored inside user_parameters
    if type(new_parameters) == type(None) : #if user clicks 'Cancel'
        return nucleus_label , cytoplasm_label, user_parameters
    else :
        segmentation_parameters.update(new_parameters)

    map_ = map_channels(segmentation_parameters)
    if type(map_) == type(None) : #User clicks Cancel 
        return nucleus_label, cytoplasm_label, user_parameters
    segmentation_parameters['reordered_shape'] = reorder_shape(segmentation_parameters['shape'], map_)
    image = reorder_image_stack(map_, segmentation_parameters['image'])

    while True : # Loop if show_segmentation 
        #Default parameters
        cyto_model_name = segmentation_parameters.setdefault('cyto_model_name', 'cyto3')
        cyto_size = segmentation_parameters.setdefault('cytoplasm diameter', 180)
        cytoplasm_channel = segmentation_parameters.setdefault('cytoplasm channel', 0)
        nucleus_model_name = segmentation_parameters.setdefault('nucleus_model_name', 'nuclei')
        nucleus_size = segmentation_parameters.setdefault('nucleus diameter', 130)
        nucleus_channel = segmentation_parameters.setdefault('nucleus channel', 0)
        other_nucleus_image = segmentation_parameters.setdefault('other_nucleus_image',None)
        path = os.getcwd()
        show_segmentation = segmentation_parameters.setdefault('show segmentation', False)
        segment_only_nuclei = segmentation_parameters.setdefault('Segment only nuclei', False)
        filename = segmentation_parameters['filename']
        available_channels = list(range(image.shape[0]))
        multichannel = segmentation_parameters.get('is_multichannel')


    #Ask user for parameters
    #if incorrect parameters --> set relaunch to True
        while True :
            layout = _segmentation_layout(
                cytoplasm_model_preset = cyto_model_name,
                cytoplasm_channel_preset= cytoplasm_channel,
                nucleus_model_preset = nucleus_model_name,
                nucleus_channel_preset= nucleus_channel,
                cyto_diameter_preset= cyto_size,
                nucleus_diameter_preset= nucleus_size,
                other_nucleus_image_preset=other_nucleus_image,
                saving_path_preset= path,
                show_segmentation_preset=show_segmentation,
                segment_only_nuclei_preset=segment_only_nuclei,
                filename_preset=filename,
                multichannel=multichannel,
            )

            event, values = prompt(layout)
            if event == 'Cancel' :
                cancel_segmentation = ask_cancel_segmentation()

                if cancel_segmentation :
                    return None, None, user_parameters
                else : 
                    continue

            #Extract parameters
            values = _cast_segmentation_parameters(values)
            do_only_nuc = values['Segment only nuclei']
            cyto_model_name = values['cyto_model_name']
            cyto_size = values['cytoplasm diameter']
            cytoplasm_channel = values['cytoplasm channel']
            nucleus_model_name = values['nucleus_model_name']
            nucleus_size = values['nucleus diameter']
            nucleus_channel = values['nucleus channel']
            other_nucleus_image = values['other_nucleus_image']
            path = values['saving path'] if values['saving path'] != '' else None
            show_segmentation = values['show segmentation']
            filename = values['filename'] if type(path) != type(None) else None
            channels = [cytoplasm_channel, nucleus_channel] if multichannel else [...,...]

            relaunch= False
            #Checking integrity of parameters
            if type(cyto_model_name) != str  and not do_only_nuc:
                sg.popup('Invalid cytoplasm model name.')
                values['cyto_model_name'] = user_parameters.setdefault('cyto_model_name', 'cyto2')
                relaunch= True
            if multichannel :
                if cytoplasm_channel not in available_channels and not do_only_nuc:
                    sg.popup('For given input image please select channel in {0}\ncytoplasm channel : {1}'.format(available_channels, cytoplasm_channel))
                    relaunch= True
                    values['cytoplasm channel'] = user_parameters.setdefault('cytoplasm channel',0)
            else :
                cytoplasm_channel = ...

            if type(cyto_size) not in [int, float] and not do_only_nuc:
                sg.popup("Incorrect cytoplasm size.")
                relaunch= True
                values['cytoplasm diameter'] = user_parameters.setdefault('diameter', 30)

            if type(nucleus_model_name) != str :
                sg.popup('Invalid nucleus model name.')
                values['nucleus_model_name'] = user_parameters.setdefault('nucleus_model_name', 'nuclei')
                relaunch= True
            
            if multichannel :
                if nucleus_channel not in available_channels :
                    sg.popup('For given input image please select channel in {0}\nnucleus channel : {1}'.format(available_channels, nucleus_channel))
                    relaunch= True
                    values['nucleus channel'] = user_parameters.setdefault('nucleus_channel', 0)
            else : 
                nucleus_channel = ...

            if type(nucleus_size) not in [int, float] :
                sg.popup("Incorrect nucleus size.")
                relaunch= True
                values['nucleus diameter'] = user_parameters.setdefault('nucleus diameter', 30)
            if other_nucleus_image != '' :
                if not os.path.isfile(other_nucleus_image) :
                    sg.popup("Nucleus image is not a file.")
                    relaunch=True
                    values['other_nucleus_image'] = None
                else :
                    try :
                        nucleus_image = open_image(other_nucleus_image)
                    except Exception as e :
                        sg.popup("Could not open image.\n{0}".format(e))
                        relaunch=True
                        values['other_nucleus_image'] = user_parameters.setdefault('other_nucleus_image', None)
                    else :
                        if nucleus_image.ndim != image.ndim - multichannel :
                            sg.popup("Nucleus image dimension missmatched. Expected same dimension as cytoplasm_image for monochannel or same dimension as cytoplasm_image -1 for multichannel\ncytoplasm dimension : {0}, nucleus dimension : {1}".format(image.ndim, nucleus_image.ndim))
                            nucleus_image = None
                            relaunch=True
                            values['other_nucleus_image'] = user_parameters.setdefault('other_nucleus_image', None)
                        
                        elif nucleus_image.shape != image[cytoplasm_channel].shape :
                            sg.popup("Nucleus image shape missmatched. Expected same shape as cytoplasm_image \ncytoplasm shape : {0}, nucleus shape : {1}".format(image[cytoplasm_channel].shape, nucleus_image.shape))
                            nucleus_image = None
                            relaunch=True
                            values['other_nucleus_image'] = user_parameters.setdefault('other_nucleus_image', None)

            else :
                nucleus_image = None

            user_parameters.update(values)
            if not relaunch : break

        #Launching segmentation
        waiting_layout = [
            [sg.Text("Running segmentation...")]
        ]
        window = sg.Window(
            title= 'small_fish',
            layout= waiting_layout,
            grab_anywhere= True,
            no_titlebar= False
        )

        window.read(timeout= 30, close= False)

        try :
            if type(path) != type(None) and filename != '':
                output_path = path + '/' + filename
                nuc_path = output_path + "_nucleus_segmentation"
                cyto_path = output_path + "_cytoplasm_segmentation"
            else :
                output_path = None
                nuc_path = None
                cyto_path = None


            cytoplasm_label, nucleus_label = cell_segmentation(
                image,
                cyto_model_name= cyto_model_name,
                cyto_diameter= cyto_size,
                nucleus_model_name= nucleus_model_name,
                nucleus_diameter= nucleus_size,
                channels=channels,
                do_only_nuc=do_only_nuc,
                external_nucleus_image = nucleus_image,
                )

        finally  : window.close()

        if show_segmentation :
            nucleus_label, cytoplasm_label = napari_show_segmentation(
                nuc_image=image[nucleus_channel] if type(nucleus_image) == type(None) else nucleus_image,
                nuc_label= nucleus_label,
                cyto_image=image[cytoplasm_channel],
                cyto_label=cytoplasm_label,
            )

            if nucleus_label.ndim == 3 : nucleus_label = np.max(nucleus_label, axis=0)
            if cytoplasm_label.ndim == 3 : cytoplasm_label = np.max(cytoplasm_label, axis=0)

            layout = [
                [sg.Text("Proceed with current segmentation ?")],
                [sg.Button("Yes"), sg.Button("No")]
            ]
            
            event, values = prompt(layout=layout, add_ok_cancel=False, add_scrollbar=False)
            if event == "No" :
                continue

        if type(output_path) != type(None) :
            
            #Get backgrounds
            nuc_proj = image[nucleus_channel]
            im_proj = image[cytoplasm_channel]
            if im_proj.ndim == 3 :
                im_proj = stack.maximum_projection(im_proj)
            if nuc_proj.ndim == 3 :
                nuc_proj = stack.maximum_projection(nuc_proj)
            
            #Call plots
            plot.plot_segmentation_boundary(nuc_proj, cytoplasm_label, nucleus_label, boundary_size=2, contrast=True, show=False, path_output=nuc_path, title= "Nucleus segmentation (blue)", remove_frame=True,)
            if not do_only_nuc : 
                plot.plot_segmentation_boundary(im_proj, cytoplasm_label, nucleus_label, boundary_size=2, contrast=True, show=False, path_output=cyto_path, title="Cytoplasm Segmentation (red)", remove_frame=True)
            plot_labels(
                nucleus_label,
                path_output=output_path + "_nucleus_label_map.png",
                show=False
                )
            if not do_only_nuc : 
                plot_labels(
                    cytoplasm_label,
                    path_output=output_path + "_cytoplasm_label_map.png",
                    show=False
                    )



        if cytoplasm_label.max() == 0 : #No cell segmented
            layout = [
            [sg.Text("No cell segmented. Proceed anyway ?")],
            [sg.Button("Yes"), sg.Button("No")]
        ]
            event, values = prompt(layout=layout, add_ok_cancel=False)
            if event == "Yes" :
                return None, None, user_parameters
        else :
            break


        
    user_parameters.update(values)
    return nucleus_label, cytoplasm_label, user_parameters

def cell_segmentation(
        image, cyto_model_name, 
        nucleus_model_name, 
        channels, cyto_diameter, 
        nucleus_diameter, 
        do_only_nuc=False,
        external_nucleus_image = None,
        ) :

    nuc_channel = channels[1]
    if not do_only_nuc : 
        cyto_channel = channels[0]
        if image[cyto_channel].ndim >= 3 :
            cyto = stack.maximum_projection(image[cyto_channel])
        else : 
            cyto = image[cyto_channel]

    if type(external_nucleus_image) != type(None) :
        nuc = external_nucleus_image
    else :
        nuc = image[nuc_channel]

    if nuc.ndim >= 3 :
        nuc = stack.maximum_projection(nuc)
    
    if not do_only_nuc :
        image = np.zeros(shape=(2,) + cyto.shape)
        image[0] = cyto
        image[1] = nuc
        image = np.moveaxis(image, source=(0,1,2), destination=(2,0,1))

    nuc_label = _segmentate_object(nuc, nucleus_model_name, nucleus_diameter, [0,0])
    if not do_only_nuc :
        cytoplasm_label = _segmentate_object(image, cyto_model_name, cyto_diameter, [1,2])
        nuc_label, cytoplasm_label = multistack.match_nuc_cell(nuc_label=nuc_label, cell_label=cytoplasm_label, single_nuc=True, cell_alone=False)
    else :
        cytoplasm_label = nuc_label

    return cytoplasm_label, nuc_label

def _segmentate_object(im, model_name, object_size_px, channels = [0,0]) :

    model = models.CellposeModel(
        gpu= use_gpu(),
        model_type= model_name,
    )

    label = model.eval(
        im,
        diameter= object_size_px,
        channels= channels,
        do_3D= False,
        )[0]
    label = np.array(label, dtype= np.int64)
    label = remove_disjoint(label)
    
    return label

def _cast_segmentation_parameters(values:dict) :

    values.setdefault('cytoplasm channel',0)
    values.setdefault('nucleus channel',0)

    if values['cyto_model_name'] == '' :
        values['cyto_model_name'] = None

    if values['nucleus_model_name'] == '' :
        values['nucleus_model_name'] = None

    try : #cytoplasm channel
        values['cytoplasm channel'] = int(values['cytoplasm channel'])
    except ValueError :
        pass

    try : #nucleus channel
        values['nucleus channel'] = int(values['nucleus channel'])
    except ValueError :
        pass

    try : #object size
        values['cytoplasm diameter'] = float(values['cytoplasm diameter'])
    except ValueError :
        pass

    try : #object size
        values['nucleus diameter'] = float(values['nucleus diameter'])
    except ValueError :
        pass
    
    return values

def remove_disjoint(image):
    """
    *CODE FROM BIG-FISH (LICENCE IN __INIT__.PY) IMPORTED TO AVOID IMPORT ERROR WHEN BIGFISH SEGMENTATION MODULE INITIALISES : try to import deprecated module for python 3.8

    For each instances with disconnected parts, keep the larger one.

    Parameters
    ----------
    image : np.ndarray, np.int, np.uint or bool
        Labelled image with shape (z, y, x) or (y, x).

    Returns
    -------
    image_cleaned : np.ndarray, np.int or np.uint
        Cleaned image with shape (z, y, x) or (y, x).

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.int32, np.int64, bool])

    # handle boolean array
    cast_to_bool = False
    if image.dtype == bool:
        cast_to_bool = bool
        image = image.astype(np.uint8)

    # initialize cleaned labels
    image_cleaned = np.zeros_like(image)

    # loop over instances
    max_label = image.max()
    for i in range(1, max_label + 1):

        # get instance mask
        mask = image == i

        # check if an instance is labelled with this value
        if mask.sum() == 0:
            continue

        # get an index for each disconnected part of the instance
        labelled_mask = label(mask)
        indices = sorted(list(set(labelled_mask.ravel())))
        if 0 in indices:
            indices = indices[1:]

        # keep the largest part of the instance
        max_area = 0
        mask_instance = None
        for j in indices:
            mask_part_j = labelled_mask == j
            area_j = mask_part_j.sum()
            if area_j > max_area:
                max_area = area_j
                mask_instance = mask_part_j

        # add instance in the final label
        image_cleaned[mask_instance] = i

    if cast_to_bool:
        image_cleaned = image_cleaned.astype(bool)

    return image_cleaned

def plot_segmentation(
        cyto_image : np.ndarray, 
        cyto_label : np.ndarray, 
        nuc_image : np.ndarray, 
        nuc_label : np.ndarray,
        path :str, 
        do_only_nuc=False
        ) :

    if nuc_image.ndim == 3 :
        nuc_image = np.max(nuc_image,axis=0)
    
    plot.plot_segmentation_boundary(
        image=nuc_image,
        nuc_label= nuc_label,
        boundary_size= 3,
        contrast=True,
        path_output=path + "_nuclei_segmentation.png",
        show=False,
    )


    if not do_only_nuc :
        if cyto_image.ndim == 3 :
            cyto_image = np.max(cyto_image,axis=0)
    
        plot.plot_segmentation_boundary(
            image=cyto_image,
            cell_label= cyto_label,
            nuc_label= nuc_label,
            boundary_size= 3,
            contrast=True,
            path_output=path + "_cytoplasm_segmentation.png",
            show=False,
        )

def plot_labels(labelled_image: np.ndarray, path_output:str = None, show= True, axis= False, close= True):
    """
    Plot a labelled image and indicate the label number at the center of each region.
    """
    stack.check_parameter(labelled_image = (np.ndarray, list), show = (bool))
    if isinstance(labelled_image, np.ndarray) : 
        stack.check_array(labelled_image, ndim= 2)
        labelled_image = [labelled_image]
    
    #Setting a colormap with background to white so all cells can be visible
    viridis = mpl.colormaps['viridis'].resampled(256)
    newcolors = viridis(np.linspace(0, 1, 256))
    white = np.array([1, 1, 1, 1])
    newcolors[0, :] = white
    newcmp = ListedColormap(newcolors)

    plt.figure(figsize= (10,10))
    rescaled_image = stack.rescale(np.array(labelled_image[0], dtype= np.int32), channel_to_stretch= 0)
    rescaled_image[rescaled_image == 0] = -100
    plot = plt.imshow(rescaled_image, cmap=newcmp)
    plot.axes.get_xaxis().set_visible(axis)
    plot.axes.get_yaxis().set_visible(axis)
    plt.tight_layout()

    for index in range(0, len(labelled_image)) :
        centroid_dict = from_label_get_centeroidscoords(labelled_image[index])
        labels = centroid_dict["label"]
        Y = centroid_dict["centroid-0"]
        X = centroid_dict["centroid-1"]
        centroids = zip(Y,X)

        for label in labels :
            y,x = next(centroids)
            y,x = round(y), round(x)
            an = plt.annotate(str(label), [round(x), round(y)])

    if not axis : plt.cla
    if show : plt.show()
    if path_output != None :
        stack.check_parameter(path_output = (str))
        plt.savefig(path_output)
    if close : plt.close()

    return plot


"""
Sub-module to handle individual spot extraction.

"""

import numpy as np
import pandas as pd
from ..interface.inoutput import write_results

def launch_spots_extraction(
        acquisition_id,
        user_parameters,
        image,
        spots,
        cluster_id,
        nucleus_label,
        cell_label,
) :
    Spots = compute_Spots(
        acquisition_id=acquisition_id,
        image=image,
        spots=spots,
        cluster_id= cluster_id,
        nucleus_label=nucleus_label,
        cell_label=cell_label,
    )

    did_output = write_results(
        Spots,
        path= user_parameters['spots_extraction_folder'],
        filename= user_parameters['spots_filename'],
        do_excel=user_parameters['do_spots_excel'],
        do_csv=user_parameters['do_spots_csv'],
        do_feather=user_parameters['do_spots_feather'],
        )
    
    if did_output : print("Individual spots extracted at {0}".format(user_parameters['spots_extraction_folder']))

def compute_Spots(
        acquisition_id : int,
        image : np.ndarray,
        spots : np.ndarray,
        cluster_id : np.ndarray,
        nucleus_label = None,
        cell_label = None,
) :

    if len(spots) == 0 :
        return pd.DataFrame()

    if type(cluster_id) == type(None) : #When user doesn't select cluster
        cluster_id = [np.NaN]*len(spots)

    index = list(zip(*spots))
    index = tuple(index)
    spot_intensities_list = list(image[index])
    if type(nucleus_label) != type(None) :
        in_nuc_list = list(nucleus_label.astype(bool)[index[-2:]]) #Only plane coordinates
    else :
        in_nuc_list = np.NaN
    if type(cell_label) != type(None) :
        cell_label_list = list(cell_label[index[-2:]]) #Only plane coordinates
    else :
        cell_label_list = np.NaN
    id_list = np.arange(len(spots))

    coord_list = list(zip(*index))

    Spots = pd.DataFrame({
        'acquisition_id' : [acquisition_id] * len(spots),
        'spot_id' : id_list,
        'intensity' : spot_intensities_list,
        'cell_label' : cell_label_list,
        'in_nucleus' : in_nuc_list,
        'coordinates' : coord_list,
        'cluster_id' : cluster_id,
    })

    return Spots
    




import numpy as np
import os
import FreeSimpleGUI as sg
from ..gui import _error_popup, _warning_popup, parameters_layout, add_header
from ..gui.prompts import input_image_prompt, prompt

class ParameterInputError(Exception) :
    """
    Raised when user inputs an incorrect parameter.
    """
    pass

class MappingError(ValueError) :
    """
    Raised when user inputs an incorrect image mapping.
    """
    def __init__(self, map_ ,*args: object) -> None:
        super().__init__(*args)
        self.map_ = map_

    def get_map(self) :
        return self.map_

def prepare_image_detection(map_, user_parameters) :
    """
    Return monochannel image for ready for spot detection; 
    if image is already monochannel, nothing happens.
    else : image is the image on which detection is performed, other_image are the other layer to show in Napari Viewer.
    """
    image = reorder_image_stack(map_, user_parameters['image'])
    assert len(image.shape) != 5 , "Time stack not supported, should never be True"
    
    if user_parameters['is_multichannel'] :
        channel_to_compute = user_parameters['channel_to_compute']
        other_image = image.copy()
        other_image = np.delete(other_image, channel_to_compute, axis=0)
        other_image = [layer for layer in other_image]
        image: np.ndarray = image[channel_to_compute]

    else :
        other_image = []

    return image, other_image

def reorder_image_stack(map_, image_stack) :
    x = (int(map_['x']),)
    y = (int(map_['y']),)
    z = (int(map_['z']),) if type(map_.get('z')) != type(None) else ()
    c = (int(map_['c']),) if type(map_.get('c')) != type(None) else ()
    t = (int(map_['t']),) if type(map_.get('t')) != type(None) else ()

    source = t+c+z+y+x

    image_stack = np.moveaxis(
        image_stack,
        source= source,
        destination= tuple(range(len(source)))
    )

    return image_stack

def map_channels(user_parameters) :
    
    image = user_parameters['image']
    is_3D_stack = user_parameters['is_3D_stack']
    is_time_stack = False
    multichannel = user_parameters['is_multichannel']

    try : 
        map_ = _auto_map_channels(is_3D_stack, is_time_stack, multichannel, image=image)
    except MappingError as e :
        sg.popup("Automatic dimension mapping went wrong. Please indicate dimensions positions in the array.")
        map_ = _ask_channel_map(image.shape, is_3D_stack, is_time_stack, multichannel, preset_map= e.get_map())

    else :
        map_ = _show_mapping(image.shape, map_, is_3D_stack, is_time_stack, multichannel,)

    return map_

def _auto_map_channels(is_3D_stack, is_time_stack, multichannel, image: np.ndarray=None, shape=None) :
    if type(shape) == type(None) :
        shape = image.shape
    reducing_list = list(shape)

    #Set the biggest dimension to y
    y_val = max(reducing_list)
    y_idx = shape.index(y_val)
    map_ = {'y' : y_idx}

    #2nd biggest set to x
    reducing_list[y_idx] = -1
    x_val = max(reducing_list)
    x_idx = reducing_list.index(x_val)
    reducing_list[y_idx] = y_val

    map_['x'] = x_idx
    reducing_list.remove(y_val)
    reducing_list.remove(x_val)

    #smaller value set to c
    if multichannel :
        c_val = min(reducing_list)
        c_idx = shape.index(c_val)
        map_['c'] = c_idx
        reducing_list.remove(c_val)

    if is_time_stack :
        t_val = reducing_list[0]
        t_idx = shape.index(t_val)
        map_['t'] = t_idx
        reducing_list.remove(t_val)
    
    if is_3D_stack :
        z_val = reducing_list[0]
        z_idx = shape.index(z_val)
        map_['z'] = z_idx

    total_channels = len(map_)
    unique_channel = len(np.unique(list(map_.values())))

    if total_channels != unique_channel : raise MappingError(map_,"{0} channel(s) are not uniquely mapped.".format(total_channels - unique_channel))

    return map_

def _ask_channel_map(shape, is_3D_stack, is_time_stack, multichannel, preset_map: dict= {}) :
    while True :
        relaunch = False
        save_preset = preset_map.copy()
        x = preset_map.setdefault('x',0)
        y = preset_map.setdefault('y',0)
        z = preset_map.setdefault('z',0)
        c = preset_map.setdefault('c',0)
        t = preset_map.setdefault('t',0)


        layout = [
            add_header("Dimensions mapping") + [sg.Text("Image shape : {0}".format(shape))]
        ]
        layout += [parameters_layout(['x','y'], default_values=[x,y])]
        if is_3D_stack : layout += [parameters_layout(['z'], default_values=[z])]
        if multichannel : layout += [parameters_layout(['c'], default_values=[c])]
        if is_time_stack : layout += [parameters_layout(['t'], default_values=[t])]

        event, preset_map = prompt(layout, add_scrollbar=False)
        if event == 'Cancel' : return save_preset

        #Check integrity
        channels_values = np.array(list(preset_map.values()), dtype= int)
        total_channels = len(preset_map)
        unique_channel = len(np.unique(channels_values))
        if total_channels != unique_channel :
            sg.popup("{0} channel(s) are not uniquely mapped.".format(total_channels - unique_channel))
            relaunch= True
        if not all(channels_values < len(shape)):
            sg.popup("Channels values out of range for image dimensions.\nPlease select dimensions from {0}".format(list(range(len(shape)))))
            relaunch= True
        if not relaunch : break

    return preset_map

def _show_mapping(shape, map_, is_3D_stack, is_time_stack, multichannel) :
    while True : 
        layout = [
            [sg.Text("Image shape : {0}".format(shape))],
            [sg.Text('Dimensions mapping was set to :')],
            [sg.Text('x : {0} \ny : {1} \nz : {2} \nc : {3} \nt : {4}'.format(
                map_['x'], map_['y'], map_.get('z'), map_.get("c"), map_.get('t')
            ))],
            [sg.Button('Change mapping')]
        ]

        event, values = prompt(layout, add_scrollbar=False)

        if event == 'Ok' :
            return map_
        elif event == 'Change mapping':
            map_ = _ask_channel_map(shape, is_3D_stack, is_time_stack, multichannel, preset_map=map_)
        elif event == 'Cancel' : 
            return None
        else : raise AssertionError('Unforseen event')


def convert_parameters_types(values:dict) :
    """
    Convert parameters from `ask_input_parameters` from strings to float, int or tuple type.
    """

    #Tuples
    tuples_dict = {
        'voxel_size' : int,
        'spot_size' : float, 
        'log_kernel_size' : float, 
        'minimum_distance' : float, 
        'deconvolution_kernel' : float,
        }
    dim = values['dim']
    if dim == 3 : dim_tuple = ('z', 'y', 'x')
    else : dim_tuple = ('y', 'x')

    for tuple_parameter, type_cast in tuples_dict.items() :
        try :
            tuple_values = tuple([type_cast(values.get(tuple_parameter + '_{0}'.format(dimension))) for dimension in dim_tuple])
        except Exception as e : #execption when str cannot be converted to float or no parameter was given.
            values[tuple_parameter] = None
        else : values[tuple_parameter] = tuple_values

    #Parameters
    int_list = ['threshold', 'channel_to_compute', 'channel_to_compute', 'min_number_of_spots', 'cluster_size','nucleus channel signal']
    float_list = ['alpha', 'beta', 'gamma', 'threshold penalty']

    for parameter in int_list :
        try :
            parameter_value = int(values[parameter])
        except Exception :
            values[parameter] = None
        else : values[parameter] = parameter_value

    for parameter in float_list :
        try :
            parameter_value = float(values[parameter])
        except Exception :
            values[parameter] = None
        else : values[parameter] = parameter_value

    return values

def check_integrity(
        values: dict, 
        do_dense_region_deconvolution,
        do_clustering, 
        multichannel,
        segmentation_done, 
        map_, 
        shape
        ):
    """
    Checks that parameters given in input by user are fit to be used for bigfish detection.
    """

    #voxel_size
    if type(values['voxel_size']) == type(None) : 
        print(values['voxel_size'])
        raise ParameterInputError('Incorrect voxel size parameter.')
    
    #detection integrity :
    if not isinstance(values['spot_size'], (tuple, list)) and not(isinstance(values['minimum_distance'], (tuple, list)) and isinstance(values['log_kernel_size'], (tuple, list))) :
       raise ParameterInputError("Either minimum_distance and 'log_kernel_size' must be correctly set\n OR 'spot_size' must be correctly set.")
    
    #Deconvolution integrity
    if do_dense_region_deconvolution :
        if not isinstance(values['alpha'], (float, int)) or not isinstance(values['beta'], (float, int)) :
            raise ParameterInputError("Incorrect alpha or beta parameters.")
        if type(values['gamma']) == type(None) and not isinstance(values['deconvolution_kernel'], (list, tuple)):
            _warning_popup('No gamma found; image will not be denoised before deconvolution.')
            values['gamma'] = 0

        if values['alpha'] > 1 or values['alpha'] < 0 :
            raise ParameterInputError("alpha must be set between 0 and 1.")

    if do_clustering :
        if not isinstance(values['min_number_of_spots'], (int)) :
            raise ParameterInputError("Incorrect min spot number parameter.")
        if not isinstance(values['cluster_size'], (int)) :
            raise ParameterInputError("Incorrect cluster size parameter.")

    #channel
    if multichannel :
        ch_len = shape[int(map_['c'])]

        if type(segmentation_done) == type(None) :
            pass
        elif segmentation_done :
            try : nuc_signal_ch = int(values['nucleus channel signal'])
            except Exception :
                raise ParameterInputError("Incorrect channel for nucleus signal measure.")
            if nuc_signal_ch > ch_len :
                raise ParameterInputError("Nucleus signal channel is out of range for image.\nPlease select from {0}".format(list(range(ch_len))))
            values['nucleus channel signal'] = nuc_signal_ch

        try :
            ch = int(values['channel_to_compute'])
        except Exception :
            raise ParameterInputError("Incorrect channel_to_compute parameter.")
        if ch >= ch_len :
            raise ParameterInputError("channel_to_compute is out of range for image.\nPlease select from {0}".format(list(range(ch_len))))
        values['channel_to_compute'] = ch

    #Spot extraction
    if not os.path.isdir(values['spots_extraction_folder']) and values['spots_extraction_folder'] != '':
        raise ParameterInputError("Incorrect spot extraction folder.")


    return values

def reorder_shape(shape, map_) :
    x = [int(map_['x']),]
    y = [int(map_['y']),]
    z = [int(map_['z']),] if type(map_.get('z')) != type(None) else []
    c = [int(map_['c']),] if type(map_.get('c')) != type(None) else []
    t = [int(map_['t']),] if type(map_.get('t')) != type(None) else []

    source = t + c + z + y + x

    new_shape = tuple(
        np.array(shape)[source]
    )

    return new_shape

def _check_segmentation_parameters(
        user_parameters,
        shape,
        is_multichannel,
) :

    available_channels = list(range(len(shape)))
    do_only_nuc = user_parameters['Segment only nuclei']
    cyto_model_name = user_parameters['cyto_model_name']
    cyto_size = user_parameters['cytoplasm diameter']
    cytoplasm_channel = user_parameters['cytoplasm channel']
    nucleus_model_name = user_parameters['nucleus_model_name']
    nucleus_size = user_parameters['nucleus diameter']
    nucleus_channel = user_parameters['nucleus channel']
   

    if type(cyto_model_name) != str  and not do_only_nuc:
        raise ParameterInputError('Invalid cytoplasm model name.')
    if cytoplasm_channel not in available_channels and not do_only_nuc and is_multichannel:
        raise ParameterInputError('For given input image please select channel in {0}\ncytoplasm channel : {1}'.format(available_channels, cytoplasm_channel))

    if type(cyto_size) not in [int, float] and not do_only_nuc:
        raise ParameterInputError("Incorrect cytoplasm size.")

    if type(nucleus_model_name) != str :
        raise ParameterInputError('Invalid nucleus model name.')
    
    if nucleus_channel not in available_channels and is_multichannel:
        raise ParameterInputError('For given input image please select channel in {0}\nnucleus channel : {1}'.format(available_channels, nucleus_channel))
    
    if type(nucleus_size) not in [int, float] :
        raise ParameterInputError("Incorrect nucleus size.")


def clean_unused_parameters_cache(user_parameters: dict) :
    """
    Clean unused parameters that were set to None in previous run.
    """
    parameters = ['alpha', 'beta', 'gamma', 'cluster_size', 'min_number_of_spots']
    for parameter in parameters :
        if parameter in user_parameters.keys() :
            if type(user_parameters[parameter]) == type(None) :
                del user_parameters[parameter]
    
    return user_parameters

def ask_input_parameters(ask_for_segmentation=True) :
    """
    Prompt user with interface allowing parameters setting for bigFish detection / deconvolution.
    """
    
    values = {}
    image_input_values = {}
    while True :
        is_3D_preset = image_input_values.setdefault('is_3D_stack', False)
        is_time_preset = image_input_values.setdefault('time stack', False)
        is_multichannel_preset = image_input_values.setdefault('is_multichannel', False)
        denseregion_preset = image_input_values.setdefault('do_dense_regions_deconvolution', False)
        do_clustering_preset = image_input_values.setdefault('do_cluster_computation', False)
        do_napari_preset = image_input_values.setdefault('show_napari_corrector', False)

        if ask_for_segmentation :
            image_input_values = input_image_prompt(
                is_3D_stack_preset=is_3D_preset,
                multichannel_preset=is_multichannel_preset,
                do_dense_regions_deconvolution_preset=None,
                do_clustering_preset= None,
                do_Napari_correction=None,
            )
        else :
            image_input_values = input_image_prompt(
                is_3D_stack_preset=is_3D_preset,
                multichannel_preset=is_multichannel_preset,
                do_dense_regions_deconvolution_preset=denseregion_preset,
                do_clustering_preset= do_clustering_preset,
                do_Napari_correction=do_napari_preset,
            )

        if type(image_input_values) == type(None) :
            return image_input_values

        if 'image' in image_input_values.keys() :
            image_input_values['shape'] = image_input_values['image'].shape 
            break


    values.update(image_input_values)
    values['dim'] = 3 if values['is_3D_stack'] else 2
    values['filename'] = os.path.basename(values['image_path'])
    
    return values


import inspect
import datetime as dt

def check_parameter(**kwargs):
    """Check dtype of the function's parameters.

    Parameters
    ----------
    kwargs : Type or Tuple[Type]
        Map of each parameter with its expected dtype.

    Returns
    -------
    _ : bool
        Assert if the array is well formatted.

    """
    # get the frame and the parameters of the function
    frame = inspect.currentframe().f_back
    _, _, _, values = inspect.getargvalues(frame)

    # compare each parameter with its expected dtype
    for arg in kwargs:
        expected_dtype = kwargs[arg]
        parameter = values[arg]
        if not isinstance(parameter, expected_dtype):
            actual = "'{0}'".format(type(parameter).__name__)
            if isinstance(expected_dtype, tuple):
                target = ["'{0}'".format(x.__name__) for x in expected_dtype]
                target = "(" + ", ".join(target) + ")"
            else:
                target = expected_dtype.__name__
            raise TypeError("Parameter {0} should be a {1}. It is a {2} "
                            "instead.".format(arg, target, actual))

    return True


def compute_anisotropy_coef(voxel_size) :
    """
    Returns tuple (anisotropy_z, anisotropy_y, 1)
    voxel_size : tuple (z,y,x).
    """

    if not isinstance(voxel_size, (tuple, list)) : raise TypeError("Expected voxel_size tuple or list")
    if len(voxel_size) == 2 : is_3D = False
    elif len(voxel_size) == 3 : is_3D = True
    else : raise ValueError("Expected 2D or 3D voxel, {0} element(s) found".format(len(voxel_size)))

    if is_3D :
        z_anisotropy = voxel_size[0] / voxel_size [2]
        xy_anisotropy = voxel_size[1] / voxel_size [2]
        return (z_anisotropy, xy_anisotropy, 1)

    else :
        return (voxel_size[0] / voxel_size[1], 1)
    

def get_datetime():
    return dt.datetime.now().strftime("%Y%m%d %H-%M-%S")


"""
Functions handling GUI update while user is interacting with software.
"""
import FreeSimpleGUI as sg

from .utils import get_elmt_from_key



def update_master_parameters(
        Master_parameter_dict:dict,
        update_dict:dict
) :
    for parameter, is_ok in Master_parameter_dict.items() :
        elmt_to_update:sg.Element = update_dict.get(parameter)
        if type(elmt_to_update) == type(None): continue
        else :
            if is_ok : 
                text:str = elmt_to_update.DisplayText.replace('Uncorrect', 'Correct')
                color = 'green'
            else : 
                text:str = elmt_to_update.DisplayText.replace('Correct', 'Uncorrect')
                color = 'gray'

            elmt_to_update.update(value=text, text_color = color)

def update_detection_tab(
        tab_elmt:sg.Tab, 
        is_multichannel, 
        is_3D, 
        do_dense_region_deconvolution,
        do_clustering,
        is_mapping_ok,
        do_segmentation,
        ) :
    
    #Acess elements
    ##Detection
    channel_to_compute = get_elmt_from_key(tab_elmt, key= 'channel_to_compute')
    voxel_size_z = get_elmt_from_key(tab_elmt, key= 'voxel_size_z')
    spot_size_z = get_elmt_from_key(tab_elmt, key= 'spot_size_z')
    log_kernel_size_z = get_elmt_from_key(tab_elmt, key= 'log_kernel_size_z')
    minimum_distance_z = get_elmt_from_key(tab_elmt, key= 'minimum_distance_z')

    ##Dense regions deconvolution
    alpha = get_elmt_from_key(tab_elmt, key= 'alpha')
    beta = get_elmt_from_key(tab_elmt, key= 'beta')
    gamma = get_elmt_from_key(tab_elmt, key= 'gamma')
    deconvolution_kernel_x = get_elmt_from_key(tab_elmt, key= 'deconvolution_kernel_x')
    deconvolution_kernel_y = get_elmt_from_key(tab_elmt, key= 'deconvolution_kernel_y')
    deconvolution_kernel_z = get_elmt_from_key(tab_elmt, key= 'deconvolution_kernel_z')
    
    #Clustering
    cluster_size = get_elmt_from_key(tab_elmt, key= 'cluster_size')
    min_number_of_spot = get_elmt_from_key(tab_elmt, key= 'min_number_of_spots')
    
    #segmentation and multichannel
    nucleus_channel_signal = get_elmt_from_key(tab_elmt, key= 'nucleus channel signal')
    
    #disable
    interactive_threshold_selector = get_elmt_from_key(tab_elmt, key= 'show_interactive_threshold_selector')

    update_dict={
        'is_3D' : is_3D,
        'is_multichannel' : is_multichannel,
        'do_dense_region_deconvolution' : do_dense_region_deconvolution,
        'do_clustering' : do_clustering,
        'always_hidden' : False,
        'is_3D&do_denseregion' : is_3D and do_dense_region_deconvolution,
        'do_segmentation' : do_segmentation and is_multichannel
    }

    list_dict={
        'is_3D' : [voxel_size_z, spot_size_z, log_kernel_size_z, minimum_distance_z, ],
        'is_multichannel' : [channel_to_compute],
        'do_dense_region_deconvolution' : [alpha,beta,gamma, deconvolution_kernel_x, deconvolution_kernel_y],
        'do_clustering' : [cluster_size, min_number_of_spot],
        'always_hidden' : [interactive_threshold_selector, ],
        'is_3D&do_denseregion' : [deconvolution_kernel_z],
        'do_segmentation' : [nucleus_channel_signal]
        
    }

    for key, enabled in update_dict.items() :
        for elmt in list_dict.get(key) :
            elmt.update(disabled=not enabled)

    tab_elmt.update(visible=is_mapping_ok)

def update_segmentation_tab(tab_elmt : sg.Tab, segmentation_correct_text : sg.Text, do_segmentation, is_multichannel, is_mapping_ok) : 
    
    #Access elements
    cytoplasm_channel_elmt = get_elmt_from_key(tab_elmt, key= 'cytoplasm channel')
    nucleus_channel_elmt = get_elmt_from_key(tab_elmt, key= 'nucleus channel')
    
    #Update values
    cytoplasm_channel_elmt.update(disabled = not is_multichannel)
    nucleus_channel_elmt.update(disabled = not is_multichannel)
    segmentation_correct_text.update(visible= do_segmentation)

    tab_elmt.update(visible=is_mapping_ok and do_segmentation)

def update_map_tab(
        tab_elmt : sg.Tab,
        is_3D,
        is_multichannel,
        last_shape,

) :
    #Acess elments
    t_element = get_elmt_from_key(tab_elmt, key= 't')
    c_element = get_elmt_from_key(tab_elmt, key= 'c')
    z_element = get_elmt_from_key(tab_elmt, key= 'z')
    automap_element = get_elmt_from_key(tab_elmt, key= 'auto-map')
    apply_element = get_elmt_from_key(tab_elmt, key= 'apply-map')

    #Update values
    t_element.update(disabled=True)
    c_element.update(disabled=not is_multichannel)
    z_element.update(disabled=not is_3D)
    automap_element.update(disabled=type(last_shape) == type(None))
    apply_element.update(disabled=type(last_shape) == type(None))

def update_output_tab(
        tab_elmt : sg.Tab,
        do_segmentation,
        output_folder,
) :
    segmentation_box = get_elmt_from_key(tab_elmt, "save segmentation")
    segmentation_box.update(disabled = not do_segmentation)
    batch_folder_text = get_elmt_from_key(tab_elmt, "batch_folder_text")
    batch_folder_text.update(value = output_folder)